#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Pytroll developers
#
# Author(s):
#
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
"""Test plugins."""

import datetime as dt
import logging
import os
import unittest
import copy
from unittest import mock
from functools import partial

import pytest
from pyresample.geometry import DynamicAreaDefinition

from trollflow2.tests.utils import TestCase
from trollflow2.launcher import read_config


yaml_test1 = """
product_list:
  something: foo
  aggregate:
    x: 2
    y: 2
  min_coverage: 5.0
  areas:
      euron1:
        areaname: euron1_in_fname
        min_coverage: 20.0
        products:
          cloud_top_height:
            productname: cloud_top_height_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image
              - format: jpg
                writer: simple_image
                fill_value: 0
            fname_pattern: "{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"

      germ:
        areaname: germ_in_fname
        fname_pattern: "{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}"
        products:
          cloudtype:
            productname: cloudtype_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image

      omerc_bb:
        areaname: omerc_bb
        output_dir: /tmp
        products:
          ct:
            productname: ct
            formats:
              - format: nc
                writer: cf
          cloud_top_height:
            productname: cloud_top_height
            formats:
              - format: tif
                writer: geotiff
"""

yaml_test_publish = """
product_list:
  something: foo
  min_coverage: 5.0
  publish_topic: /raster/
  extra_metadata:
    processing_center: SMHI
  areas:
      euron1:
        areaname: euron1_in_fname
        min_coverage: 20.0
        products:
          cloud_top_height:
            productname: cloud_top_height_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image
                filename: /tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png
                dispatch:
                  - hostname: ftp.important_client.com
                    scheme: ftp
                    path: "/somewhere/{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"
              - format: jpg
                writer: simple_image
                fill_value: 0
                filename: /tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.jpg
            fname_pattern: "{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"

      germ:
        areaname: germ_in_fname
        fname_pattern: "{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}"
        products:
          cloudtype:
            productname: cloudtype_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image
                filename: /tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_germ_in_fname_ct.png

      omerc_bb:
        areaname: omerc_bb
        output_dir: /tmp
        products:
          ct:
            productname: ct
            formats:
              - format: nc
                writer: cf
                filename: /tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_omerc_bb_in_fname_ct.png
          cloud_top_height:
            productname: cloud_top_height
            formats:
              - format: tif
                writer: geotiff
                filename: /tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_omerc_bb_in_fname_ctth.png
      null:
          areaname: satproj
          fname_pattern: "{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}"
          output_dir: /tmp
          products:
            ("chl_nn", "chl_oc4me", "trsp", "tsm_nn", "iop_nn", "mask", "latitude", "longitude"):
              min_sunlight_coverage: 10
              productname: sat_coast
              publish_topic: /nc/2C/olcil2
              formats:
                - format: nc
                  writer: cf
                  encoding:
                    latitude: {'dtype': 'int32', 'scale_factor': 1.0e-6, '_FillValue': -200000000, 'zlib': true}
                    longitude: {'dtype': 'int32', 'scale_factor': 1.0e-6, '_FillValue': -200000000, 'zlib': true}
"""

yaml_test3 = """
product_list:
  something: foo
  min_coverage: 5.0
  areas:
      euron1:
        areaname: euron1_in_fname
        min_coverage: 20.0
        products:
          green_snow:
            productname: green_snow
            formats:
              - format: tif
                writer: geotiff
"""

yaml_test_save = """
product_list:
  output_dir: &output_dir
    /tmp/satnfs/polar_out/pps2018/direct_readout/
  publish_topic: /NWC-CF/L3
  use_extern_calib: false
  fname_pattern: &fname
    "{platform_name}_{start_time:%Y%m%d_%H%M}_{areaname}_{productname}.{format}"
  formats: &formats
    - format: tif
      writer: geotiff
    - format: nc
      writer: cf
  areas:
      euron1:
        areaname: euron1_in_fname
        min_coverage: 20.0
        products:
          cloud_top_height:
            productname: cloud_top_height_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image
                dispatch:
                  - hostname: ftp.important_client.com
                    scheme: ftp
                    path: "/somewhere/{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"
              - format: jpg
                writer: simple_image
                fill_value: 0
            fname_pattern: "{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"
          ("ct", "ctth"):
            productname: ct_and_ctth
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: nc
                writer: cf

      germ:
        areaname: germ_in_fname
        use_tmp_file: True
        fname_pattern: "{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}"
        products:
          cloudtype:
            productname: cloudtype_in_fname
            output_dir: /tmp/satdmz/pps/www/latest_2018/{orig_platform_name}/
            formats:
              - format: png
                writer: simple_image

      omerc_bb:
        areaname: omerc_bb
        output_dir: /tmp
        products:
          ct:
            productname: ct
            formats:
              - format: nc
                writer: cf
          cloud_top_height:
            productname: cloud_top_height
            resolution: 500
            formats:
              - format: tif
                writer: geotiff
"""

yaml_test_null_area = """
product_list:
  areas:
      null:
        areaname: foo
        products:
          cloud_top_height:
            productname: ctth
            output_dir: /tmp/
            formats:
              - format: png
                writer: simple_image
            fname_pattern: "{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}"
"""

input_mda = {'orig_platform_name': 'noaa15', 'orbit_number': 7993,
             'start_time': dt.datetime(2019, 2, 17, 6, 0, 11, 100000), 'stfrac': 1,
             'end_time': dt.datetime(2019, 2, 17, 6, 15, 10, 400000), 'etfrac': 4, 'status': 'OK',
             'format': 'CF', 'data_processing_level': '2', 'orbit': 7993, 'module': 'ppsMakePhysiography',
             'platform_name': 'NOAA-15', 'pps_version': 'v2018', 'file_was_already_processed': False,
             'dataset': [{'uri': '/home/a001673/data/satellite/test_trollflow2/S_NWC_CMA_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc',  # noqa
                          'uid': 'S_NWC_CMA_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc'},
                         {'uri': '/home/a001673/data/satellite/test_trollflow2/S_NWC_CTTH_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc',  # noqa
                          'uid': 'S_NWC_CTTH_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc'},
                         {'uri': '/home/a001673/data/satellite/test_trollflow2/S_NWC_CT_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc',  # noqa
                          'uid': 'S_NWC_CT_noaa15_07993_20190217T0600111Z_20190217T0615104Z.nc'}],
             'sensor': ['avhrr']}

YAML_FILE_PUBLISHER = """
!!python/object:trollflow2.plugins.FilePublisher {port: 40002, nameservers: [localhost]}
"""

SCENE_START_TIME = dt.datetime.utcnow()
SCENE_END_TIME = SCENE_START_TIME + dt.timedelta(minutes=15)
JOB_INPUT_MDA_START_TIME = SCENE_START_TIME + dt.timedelta(seconds=10)


class TestSaveDatasets(TestCase):
    """Test case for saving datasets."""

    def test_prepared_filename(self):
        """Test the `prepared_filename` context."""
        from trollflow2.plugins import prepared_filename
        tst_file = 'hi.png'

        renames = {}
        fmat = {'fname_pattern': tst_file}
        with prepared_filename(fmat, renames) as filename:
            pass
        self.assertEqual(filename, tst_file)
        self.assertEqual(len(renames), 0)

        renames = {}
        fmat = {'use_tmp_file': False, 'fname_pattern': tst_file}
        with prepared_filename(fmat, renames) as filename:
            pass
        self.assertEqual(filename, tst_file)
        self.assertEqual(len(renames), 0)

        renames = {}
        fmat = {'use_tmp_file': True, 'fname_pattern': tst_file}
        with prepared_filename(fmat, renames) as filename:
            pass
        self.assertTrue(filename.startswith, 'tmp')
        self.assertEqual(len(renames), 1)
        self.assertEqual(list(renames.values())[0], tst_file)

        renames = {}
        fmat = {'use_tmp_file': True, 'fname_pattern': tst_file}
        try:
            with prepared_filename(fmat, renames) as filename:
                raise KeyError('Oh no!')
        except KeyError:
            pass
        self.assertTrue(filename.startswith, 'tmp')
        self.assertEqual(len(renames), 0)

        tst_dir = os.path.normpath('/tmp/bleh')
        renames = {}
        fmat = {'use_tmp_file': True, 'fname_pattern': tst_file, 'output_dir': tst_dir}
        with prepared_filename(fmat, renames) as filename:
            pass
        self.assertTrue(filename.startswith, 'tmp')
        self.assertEqual(len(renames), 1)
        self.assertEqual(list(renames.values())[0], os.path.join(tst_dir, tst_file))

        self.assertTrue(os.path.exists(tst_dir))
        os.rmdir(tst_dir)

    def test_prepare_filename_and_directory(self):
        """Test filename composition and directory creation."""
        from trollflow2.plugins import _prepare_filename_and_directory
        tst_file = 'goes_{name}.png'
        tst_dir = os.path.normpath('/tmp/bleh/{service}')
        fmat = {'use_tmp_file': True, 'fname_pattern': tst_file, 'output_dir': tst_dir,
                'name': 'mooh', 'service': 'cow'}
        directory, filename = _prepare_filename_and_directory(fmat)
        self.assertEqual(filename, os.path.normpath('/tmp/bleh/cow/goes_mooh.png'))
        self.assertTrue(os.path.exists(directory))
        os.rmdir(directory)

    def test_get_temp_filename(self):
        """Test temp filename generation."""
        from trollflow2.plugins import _get_temp_filename

        # Test uniqueness
        class FakeFO():
            def __init__(self, name):
                self.name = name

            def close(self):
                pass

        tst_dir = ''
        with mock.patch('trollflow2.plugins.NamedTemporaryFile') as ntf:
            ntf.side_effect = [FakeFO('cu'), FakeFO('cu'), FakeFO('mb'), FakeFO('er')]
            names = []
            for _i_ in range(3):
                names.append(_get_temp_filename(tst_dir, names))
            self.assertEqual(''.join(names), 'cumber')

    def test_save_datasets(self):
        """Test saving datasets."""
        self.maxDiff = None
        from trollflow2.plugins import save_datasets, DEFAULT

        the_queue = mock.MagicMock()
        job = _create_job_for_save_datasets()
        job['produced_files'] = the_queue
        with mock.patch('trollflow2.plugins.compute_writer_results'),\
                mock.patch('trollflow2.plugins.DataQuery') as dsid,\
                mock.patch('os.rename') as rename:
            save_datasets(job)
            expected_sd = [mock.call(dsid.return_value, compute=False,
                                     filename=os.path.join('/tmp', 'satdmz', 'pps', 'www', 'latest_2018',
                                                           'NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png'),
                                     format='png', writer='simple_image'),
                           mock.call(dsid.return_value, compute=False,
                                     filename=os.path.join('/tmp', 'satdmz', 'pps', 'www', 'latest_2018',
                                                           'NOAA-15_20190217_0600_euron1_in_fname_ctth_static.jpg'),
                                     fill_value=0, format='jpg', writer='simple_image'),
                           mock.call(dsid.return_value, compute=False,
                                     filename=os.path.join('/tmp', 'NOAA-15_20190217_0600_omerc_bb_ct.nc'),
                                     format='nc', writer='cf'),
                           mock.call(dsid.return_value, compute=False,
                                     filename=os.path.join('/tmp',
                                                           'NOAA-15_20190217_0600_omerc_bb_cloud_top_height.tif'),
                                     format='tif', writer='geotiff')
                           ]
            expected_sds = [mock.call(datasets=[dsid.return_value, dsid.return_value], compute=False,
                                      filename=os.path.join('/tmp', 'satdmz', 'pps', 'www', 'latest_2018',
                                                            'NOAA-15_20190217_0600_euron1_in_fname_ct_and_ctth.nc'),
                                      writer='cf')]
            expected_dsid = [mock.call(name='cloud_top_height', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='cloud_top_height', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='ct', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='ctth', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='cloudtype', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='ct', resolution=DEFAULT, modifiers=DEFAULT),
                             mock.call(name='cloud_top_height', resolution=500, modifiers=DEFAULT)
                             ]

            sd_calls = (job['resampled_scenes']['euron1'].save_dataset.mock_calls
                        + job['resampled_scenes']['omerc_bb'].save_dataset.mock_calls)
            for sd, esd in zip(sd_calls, expected_sd):
                self.assertEqual(sd, esd)
            sds_calls = job['resampled_scenes']['euron1'].save_datasets.mock_calls
            for sds, esds in zip(sds_calls, expected_sds):
                self.assertDictEqual(sds[2], esds[2])
            args, kwargs = job['resampled_scenes']['germ'].save_dataset.call_args_list[0]
            self.assertTrue(os.path.basename(kwargs['filename']).startswith('tmp'))
            for ds, eds in zip(dsid.mock_calls, expected_dsid):
                self.assertEqual(ds, eds)
            rename.assert_called_once()

        dexpected = {
            'output_dir': '/tmp/satnfs/polar_out/pps2018/direct_readout/',
            'publish_topic': '/NWC-CF/L3',
            'use_extern_calib': False,
            'fname_pattern': "{platform_name}_{start_time:%Y%m%d_%H%M}_{areaname}_{productname}.{format}",
            'formats': [
                {
                    'format': 'tif',
                    'writer': 'geotiff'
                },
                {
                    'format': 'nc',
                    'writer': 'cf'
                }],
            'areas': {
                'euron1': {
                    'areaname': 'euron1_in_fname',
                    'min_coverage': 20.0,
                    'products': {
                        'cloud_top_height': {
                            'fname_pattern': '{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}',  # noqa
                            'formats':
                            [{
                                'filename': '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png',  # noqa
                                'format': 'png',
                                'writer': 'simple_image',
                                'dispatch': [{
                                    'hostname': 'ftp.important_client.com',
                                    'scheme': 'ftp',
                                    'path': '/somewhere/{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth_static.{format}',  # noqa
                                 }],
                             },
                             {
                                 'filename':
                                 '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.jpg',  # noqa
                                 'fill_value':
                                 0,
                                 'format':
                                 'jpg',
                                 'writer':
                                 'simple_image'
                             }],
                            'output_dir':
                            '/tmp/satdmz/pps/www/latest_2018/',
                            'productname':
                            'cloud_top_height_in_fname'
                        },
                        ('ct', 'ctth'): {
                            'formats':
                            [{
                                'filename': '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ct_and_ctth.nc',  # noqa
                                'format': 'nc',
                                'writer': 'cf'
                            }],
                            'output_dir': '/tmp/satdmz/pps/www/latest_2018/',
                            'productname': 'ct_and_ctth'
                        },
                    }
                },
                'germ': {
                    'areaname':
                    'germ_in_fname',
                    'use_tmp_file':
                    True,
                    'fname_pattern':
                    '{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}',
                    'products': {
                        'cloudtype': {
                            'formats': [{
                                'filename':
                                '/tmp/satdmz/pps/www/latest_2018/noaa15/20190217_0600_germ_in_fname_cloudtype_in_fname.png',  # noqa
                                'format':
                                'png',
                                'writer':
                                'simple_image'
                            }],
                            'output_dir':
                            '/tmp/satdmz/pps/www/latest_2018/{orig_platform_name}/',
                            'productname':
                            'cloudtype_in_fname'
                        }
                    }
                },
                'omerc_bb': {
                    'areaname': 'omerc_bb',
                    'output_dir': '/tmp',
                    'products': {
                        'cloud_top_height': {
                            'formats': [{
                                'filename':
                                '/tmp/NOAA-15_20190217_0600_omerc_bb_cloud_top_height.tif',
                                'format':
                                'tif',
                                'writer':
                                'geotiff'
                            }],
                            'productname':
                            'cloud_top_height',
                            'resolution':
                            500
                        },
                        'ct': {
                            'formats': [{
                                'filename':
                                '/tmp/NOAA-15_20190217_0600_omerc_bb_ct.nc',
                                'format':
                                'nc',
                                'writer':
                                'cf'
                            }],
                            'productname':
                            'ct'
                        }
                    }
                }
            }
        }
        self.assertDictEqual(dexpected, job['product_list']['product_list'])

        filenames = ['/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png',
                     '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.jpg',
                     '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ct_and_ctth.nc',
                     '/tmp/satdmz/pps/www/latest_2018/noaa15/20190217_0600_germ_in_fname_cloudtype_in_fname.png',
                     '/tmp/NOAA-15_20190217_0600_omerc_bb_ct.nc',
                     '/tmp/NOAA-15_20190217_0600_omerc_bb_cloud_top_height.tif']
        for fname, efname in zip(the_queue.put.mock_calls, filenames):
            self.assertEqual(fname, mock.call(efname))

    def test_save_datasets_eager(self):
        """Test saving datasets in eager manner."""
        from trollflow2.plugins import save_datasets

        job = _create_job_for_save_datasets()
        job['product_list']['product_list']['eager_writing'] = True
        with mock.patch('trollflow2.plugins.compute_writer_results') as compute_writer_results,\
                mock.patch('trollflow2.plugins.DataQuery'),\
                mock.patch('os.rename'):
            save_datasets(job)
            sd_calls = (job['resampled_scenes']['euron1'].save_dataset.mock_calls
                        + job['resampled_scenes']['omerc_bb'].save_dataset.mock_calls)
            for sd in sd_calls:
                assert "compute=True" in str(sd)
            sds_calls = job['resampled_scenes']['euron1'].save_datasets.mock_calls
            for sds in sds_calls:
                assert "compute=True" in str(sds)
            compute_writer_results.assert_not_called()

    def test_pop_unknown_args(self):
        """Test pop unknown kwargs."""
        from trollflow2.plugins import save_datasets
        job = _create_job_for_save_datasets()

        product_list = {
            "fname_pattern": "name.tif",
            "use_tmp_file": True,
            "staging_zone": "értékesítési szakember",
            "areas": {
                "euron1": {
                    "products": {
                        "IR_108": {
                            "productname": "IR108",
                            "formats": [
                                {"writer": "ninjogeotiff",
                                 "ChannelID": 0,
                                 "DataType": 0,
                                 "PhysicUnit": "no",
                                 "PhysicValue": "yes",
                                 "SatelliteNameID": 0,
                                 "output_dir": "örülök, hogy megismerhetem",
                                 "fname_pattern": "viszontlátásra",
                                 "dispatch": {},
                                 "use_tmp_file": False,
                                 "staging_zone": "értékesítési szakember",
                                 }
                            ]
                        }
                    }
                }
            }
        }
        job["product_list"] = {"product_list": product_list}

        with mock.patch('trollflow2.plugins.compute_writer_results'), \
                mock.patch("os.rename"):
            save_datasets(job)

        assert "PhysicUnit" in job["resampled_scenes"]["euron1"].mock_calls[0].kwargs.keys()
        for absent in {"use_tmp_file", "staging_zone", "output_dir",
                       "fname_pattern", "dispatch"}:
            assert absent not in job["resampled_scenes"]["euron1"].mock_calls[0].kwargs.keys()


def _create_job_for_save_datasets():
    from yaml import UnsafeLoader
    job = {}
    job['input_mda'] = input_mda
    job['product_list'] = {
        'product_list': read_config(raw_string=yaml_test_save, Loader=UnsafeLoader)['product_list'],
    }
    job['resampled_scenes'] = {}
    job['produced_files'] = mock.Mock()
    for area in job['product_list']['product_list']['areas']:
        job['resampled_scenes'][area] = mock.Mock()
    return job


def test_use_staging_zone_no_tmpfile():
    """Test `prepared_filename` context with staging zone.

    Test that when staging_zone is set, the output file is created in this
    directory first, before being moved to output_dir.
    """
    from trollflow2.plugins import prepared_filename
    tst_file = "trappedinaunittest.tif"

    renames = {}
    fmat = {"use_tmp_file": False, "fname_pattern": tst_file,
            "staging_zone": "/dummy/abcd"}
    with prepared_filename(fmat, renames) as filename:
        pass
    assert filename != tst_file
    assert filename.endswith(tst_file)
    assert len(renames) == 1
    assert next(iter(renames.values())) == tst_file
    assert next(iter(renames.keys())).startswith("/dummy/abcd/")


def test_use_staging_zone_tmpfile(tmp_path):
    """Test `prepared_filename` context with staging zone and tmpfile.

    Test that when both staging zone and tmpfile are set, an output file
    with a temporary name is created in the staging zone directory.
    """
    from trollflow2.plugins import prepared_filename
    tst_file = "stilltrappedinaunittest.tif"

    renames = {}
    fmat = {"use_tmp_file": True, "fname_pattern": tst_file,
            "staging_zone": os.fspath(tmp_path)}
    with prepared_filename(fmat, renames) as filename:
        pass
    assert filename != tst_file
    assert not filename.endswith(tst_file)
    assert len(renames) == 1
    assert next(iter(renames.values())) == tst_file
    assert next(iter(renames.keys())).startswith(os.fspath(tmp_path))


class TestCreateScene(TestCase):
    """Test case for creating a scene."""

    def test_create_scene(self):
        """Test making a scene."""
        from trollflow2.plugins import create_scene
        from satpy.version import version as satpy_version
        if not isinstance(satpy_version, str):
            # trollflow2 mocks all missing imports
            import trollflow2.plugins
            trollflow2.plugins.satpy_version = satpy_version = "0.26.0"
        with mock.patch("trollflow2.plugins.Scene", autospec=True) as scene:
            scene.return_value = "foo"
            job = {"input_filenames": "bar", "product_list": {}}
            create_scene(job)
            self.assertEqual(job["scene"], "foo")
            if satpy_version <= "0.25.1":
                scene.assert_called_with(filenames='bar', reader=None,
                                         reader_kwargs=None, ppp_config_dir=None)
            else:
                scene.assert_called_with(filenames='bar', reader=None,
                                         reader_kwargs=None)
            job = {"input_filenames": "bar",
                   "product_list": {"product_list": {"reader": "baz"}}}
            create_scene(job)
            if satpy_version <= "0.25.1":
                scene.assert_called_with(filenames='bar', reader='baz',
                                         reader_kwargs=None, ppp_config_dir=None)
            else:
                scene.assert_called_with(filenames='bar', reader='baz',
                                         reader_kwargs=None)


class TestLoadComposites(TestCase):
    """Test case for loading composites."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)

    def test_load_composites(self):
        """Test loading composites."""
        from trollflow2.plugins import load_composites, DEFAULT
        scn = _get_mocked_scene_with_properties()
        job = {"product_list": self.product_list, "scene": scn}
        load_composites(job)
        scn.load.assert_called_with({'ct', 'cloudtype', 'cloud_top_height'}, resolution=DEFAULT, generate=False)

    def test_load_composites_with_config(self):
        """Test loading composites with a config."""
        from trollflow2.plugins import load_composites
        scn = _get_mocked_scene_with_properties()
        self.product_list['product_list']['resolution'] = 1000
        self.product_list['product_list']['delay_composites'] = False
        job = {"product_list": self.product_list, "scene": scn}
        load_composites(job)
        scn.load.assert_called_with({'ct', 'cloudtype', 'cloud_top_height'}, resolution=1000, generate=True)

    def test_load_composites_with_different_resolutions(self):
        """Test loading composites with different resolutions."""
        from trollflow2.plugins import load_composites
        scn = _get_mocked_scene_with_properties()
        self.product_list['product_list']['resolution'] = 1000
        self.product_list['product_list']['areas']['euron1']['resolution'] = 500
        self.product_list['product_list']['delay_composites'] = False
        job = {"product_list": self.product_list, "scene": scn}
        load_composites(job)
        scn.load.assert_any_call({'cloudtype', 'ct', 'cloud_top_height'}, resolution=1000, generate=True)
        scn.load.assert_any_call({'cloud_top_height'}, resolution=500, generate=True)

    def test_load_composites_with_custom_args(self):
        """Test loading with arbitrary additional arguments."""
        from trollflow2.plugins import load_composites, DEFAULT
        scn = _get_mocked_scene_with_properties()
        self.product_list['product_list']['scene_load_kwargs'] = {"upper_right_corner": "NE"}
        job = {"product_list": self.product_list, "scene": scn}
        load_composites(job)
        scn.load.assert_called_with(
            {'ct', 'cloudtype', 'cloud_top_height'},
            resolution=DEFAULT, generate=False, upper_right_corner="NE")


class TestAggregate(TestCase):
    """Test case for aggregating."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)

    def test_aggregate_returns_aggregated_scene(self):
        """Test aggregating."""
        from trollflow2.plugins import aggregate
        scn = _get_mocked_scene_with_properties()
        assert 'aggregate' in self.product_list['product_list']
        job = {"scene": scn, "product_list": self.product_list}
        aggregate(job)
        assert job['scene'] is scn.aggregate.return_value

    def test_aggregate_is_called_with_right_params(self):
        """Test aggregating."""
        from trollflow2.plugins import aggregate
        scn = _get_mocked_scene_with_properties()
        assert 'aggregate' in self.product_list['product_list']
        self.product_list['product_list']['aggregate'] = dict(x=4, y=4)
        job = {"scene": scn, "product_list": self.product_list}
        aggregate(job)
        scn.aggregate.assert_called_once_with(x=4, y=4)

    def test_aggregate_returns_original_scene_when_not_needed(self):
        """Test aggregating."""
        from trollflow2.plugins import aggregate
        scn = _get_mocked_scene_with_properties()
        del self.product_list['product_list']['aggregate']
        job = {"scene": scn, "product_list": self.product_list}
        aggregate(job)
        assert job['scene'] is scn


class TestResample(TestCase):
    """Test case for resampling."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)

    def test_resample(self):
        """Test resampling."""
        from trollflow2.plugins import resample
        scn = _get_mocked_scene_with_properties()
        scn.resample.return_value = "foo"
        job = {"scene": scn, "product_list": self.product_list}
        resample(job)
        self.assertTrue(mock.call('omerc_bb',
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        self.assertTrue(mock.call('germ',
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        self.assertTrue(mock.call('euron1',
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        self.assertTrue("resampled_scenes" in job)
        for area in ["omerc_bb", "germ", "euron1"]:
            self.assertTrue(area in job["resampled_scenes"])
            self.assertTrue(job["resampled_scenes"][area] == "foo")

        prod_list = self.product_list.copy()
        prod_list["common"] = {"resampler": "bilinear"}
        prod_list["product_list"]['areas']["euron1"]["reduce_data"] = False
        job = {"product_list": prod_list, "scene": scn}
        resample(job)
        self.assertTrue(mock.call('euron1',
                                  radius_of_influence=None,
                                  resampler="bilinear",
                                  reduce_data=False,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)

    def test_resample_satproj(self):
        """Test keeping the satellite projection."""
        from trollflow2.plugins import resample
        scn = mock.MagicMock()
        scn.resample.return_value = "foo"
        job = {"scene": scn, "product_list": self.product_list.copy()}
        job['product_list']['product_list']['areas']['None'] = job['product_list']['product_list']['areas']['germ']
        del job['product_list']['product_list']['areas']['germ']
        resample(job)
        self.assertTrue(mock.call('omerc_bb',
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        self.assertTrue(mock.call('euron1',
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        self.assertTrue(job['resampled_scenes']['None'] is scn)
        self.assertTrue("resampled_scenes" in job)
        for area in ["omerc_bb", "euron1"]:
            self.assertTrue(area in job["resampled_scenes"])
            self.assertTrue(job["resampled_scenes"][area] == "foo")

    def test_minmax_area(self):
        """Test the min and max areas."""
        from trollflow2.plugins import resample
        scn = _get_mocked_scene_with_properties()
        scn.resample.return_value = "foo"
        product_list = self.product_list.copy()
        product_list['product_list']['areas']['None'] = product_list['product_list']['areas']['germ']
        product_list['product_list']['areas']['None']['use_min_area'] = True
        del product_list['product_list']['areas']['germ']
        del product_list['product_list']['areas']['omerc_bb']
        del product_list['product_list']['areas']['euron1']
        job = {"scene": scn, "product_list": product_list.copy()}
        resample(job)
        self.assertTrue(mock.call(scn.min_area(),
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)
        del product_list['product_list']['areas']['None']['use_min_area']
        product_list['product_list']['areas']['None']['use_max_area'] = True
        job = {"scene": scn, "product_list": product_list.copy()}
        resample(job)
        self.assertTrue(mock.call(scn.max_area(),
                                  radius_of_influence=None,
                                  resampler="nearest",
                                  reduce_data=True,
                                  cache_dir=None,
                                  mask_area=False,
                                  epsilon=0.0) in
                        scn.resample.mock_calls)


class TestResampleNullArea(TestCase):
    """Test case for resampling."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test_null_area, Loader=UnsafeLoader)

    def test_resample_null_area(self):
        """Test handling a `None` area in resampling."""
        from trollflow2.plugins import resample
        scn = _get_mocked_scene_with_properties()
        product_list = self.product_list.copy()
        job = {"scene": scn, "product_list": product_list.copy()}
        # The composites have been generated
        scn.keys.return_value = ['abc']
        scn.wishlist = {'abc'}
        resample(job)
        scn.load.assert_not_called()
        # The composites have not been generated
        scn.keys.return_value = ['a', 'b', 'c']
        scn.wishlist = {'abc'}
        resample(job)
        self.assertTrue(mock.call({'abc'}, generate=True) in
                        scn.load.mock_calls)

    def test_resample_native_null_area(self):
        """Test using `native` resampler with `None` area."""
        from trollflow2.plugins import resample
        scn = _get_mocked_scene_with_properties()
        product_list = self.product_list.copy()
        product_list["common"] = {"resampler": "native"}
        job = {"scene": scn, "product_list": product_list.copy()}
        # The composites have been generated
        scn.keys.return_value = ['abc']
        scn.wishlist = {'abc'}
        resample(job)
        self.assertTrue("resampler='native'" in
                        str(scn.resample.mock_calls))


class TestSunlightCovers(TestCase):
    """Test the sunlight coverage."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)
        self.input_mda = {"platform_name": "NOAA-15",
                          "sensor": "avhrr-3",
                          "start_time": dt.datetime(2019, 4, 7, 20, 52),
                          "end_time": dt.datetime(2019, 4, 7, 20, 58),
                          }

    def test_coverage(self):
        """Test sunlight coverage."""
        from trollflow2.plugins import _get_sunlight_coverage
        import numpy as np
        with mock.patch('trollflow2.plugins.AreaDefBoundary') as area_def_boundary, \
                mock.patch('trollflow2.plugins.Boundary') as boundary, \
                mock.patch('trollflow2.plugins.get_twilight_poly'), \
                mock.patch('trollflow2.plugins.get_area_def'), \
                mock.patch('trollflow2.plugins.get_geostationary_bounding_box'):

            area_def_boundary.return_value.contour_poly.intersection.return_value.area.return_value = 0.02
            boundary.return_value.contour_poly.intersection.return_value.area.return_value = 0.02
            area_def_boundary.return_value.contour_poly.area.return_value = 0.2
            start_time = dt.datetime(2019, 4, 7, 20, 8)
            adef = mock.MagicMock(proj_dict={'proj': 'stere'})
            res = _get_sunlight_coverage(adef, start_time)
            np.testing.assert_allclose(res, 0.1)
            boundary.assert_not_called()
            adef = mock.MagicMock(proj_dict={'proj': 'geos'})
            res = _get_sunlight_coverage(adef, start_time)
            boundary.assert_called()


class TestGetProductAreaDef(TestCase):
    """Test case for finding area definition for a product."""

    def test_get_product_area_def(self):
        """Test _get_product_area_def()."""
        from trollflow2.plugins import _get_product_area_def
        # scn = mock.MagicMock()
        # scn.__getitem__.side_effect = KeyError

        # No area nor product
        scn = dict([])
        job = {'scene': scn}
        area = 'area'
        product = 'product'
        res = _get_product_area_def(job, area, product)
        self.assertIsNone(res)

        # Area not in the scene, take area def from the available first dataset
        adef = mock.MagicMock()
        prod = mock.MagicMock()
        prod.attrs.__getitem__.return_value = adef
        scn['1'] = prod
        job = {'scene': scn}
        res = _get_product_area_def(job, area, product)
        self.assertTrue(res is adef)
        prod.attrs.__getitem__.assert_called_once()

        # Area from the un-resampled scene
        adef = mock.MagicMock()
        prod = mock.MagicMock()
        prod.attrs.__getitem__.return_value = adef
        prod2 = mock.MagicMock()
        prod2.attrs.__getitem__.return_value = None
        scn = {area: prod, '1': prod2}
        job = {'scene': scn}
        res = _get_product_area_def(job, area, product)
        self.assertTrue(res is adef)
        prod.attrs.__getitem__.assert_called_once()
        prod2.attrs.__getitem__.assert_not_called()
        # Product is a tuple
        res = _get_product_area_def(job, area, (product, 'foo'))
        self.assertTrue(res is adef)

        # Area from a resampled scene
        adef = mock.MagicMock()
        prod = mock.MagicMock()
        prod.attrs.__getitem__.return_value = adef
        prod2 = mock.MagicMock()
        prod2.attrs.__getitem__.return_value = None
        scn = {area: prod, '1': prod2}
        job = {'resampled_scenes': {area: scn}}
        res = _get_product_area_def(job, area, product)
        self.assertTrue(res is adef)
        prod.attrs.__getitem__.assert_called_once()
        prod2.attrs.__getitem__.assert_not_called()
        # Product is a tuple
        res = _get_product_area_def(job, area, (product, 'foo'))
        self.assertTrue(res is adef)


class TestCheckSunlightCoverage(TestCase):
    """Test case for sunlight coverage."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test3, Loader=UnsafeLoader)
        self.input_mda = {"platform_name": "NOAA-15",
                          "sensor": "avhrr-3",
                          "start_time": dt.datetime(2019, 1, 19, 11),
                          "end_time": dt.datetime(2019, 1, 19, 12),
                          'not_changed': True,
                          }

    def test_metadata_is_read_from_scene(self):
        """Test that the scene and message metadata are merged correctly."""
        from trollflow2.plugins import check_sunlight_coverage

        with mock.patch('trollflow2.plugins.Pass') as ts_pass,\
                mock.patch('trollflow2.plugins.get_twilight_poly'),\
                mock.patch('trollflow2.plugins.get_area_def'),\
                mock.patch("trollflow2.plugins._get_sunlight_coverage") as _get_sunlight_coverage:
            _get_sunlight_coverage.return_value = .3
            scene = _get_mocked_scene_with_properties()
            job = {"scene": scene, "product_list": self.product_list.copy(),
                   "input_mda": {"platform_name": "platform"}}
            job['product_list']['product_list']['sunlight_coverage'] = {'min': 10, 'max': 40, 'check_pass': True}
            check_sunlight_coverage(job)
            ts_pass.assert_called_with(job["input_mda"]["platform_name"], scene.start_time, scene.end_time,
                                       instrument=list(scene.sensor_names)[0])

    def test_fully_sunlit_scene_returns_full_coverage(self):
        """Test that a fully sunlit scene returns 100% coverage."""
        from trollflow2.plugins import check_sunlight_coverage
        from pyresample.spherical import SphPolygon
        import numpy as np
        with mock.patch('trollflow2.plugins.Pass') as tst_pass,\
                mock.patch('trollflow2.plugins.get_twilight_poly') as twilight:
            tst_pass.return_value.boundary.contour_poly = SphPolygon(np.array([(0, 0), (0, 90), (45, 0)]))
            twilight.return_value = SphPolygon(np.array([(0, 0), (0, 90), (90, 0)]))
            scene = _get_mocked_scene_with_properties()
            job = {"scene": scene, "product_list": self.product_list.copy(),
                   "input_mda": {"platform_name": "platform"}}
            job['product_list']['product_list']['sunlight_coverage'] = {'min': 10, 'max': 40, 'check_pass': True}
            check_sunlight_coverage(job)
            assert job['product_list']['product_list']['areas']['euron1']['area_sunlight_coverage_percent'] == 100

    def test_product_not_loaded(self):
        """Test that product isn't loaded when sunlight coverage is too low."""
        from trollflow2.plugins import check_sunlight_coverage
        from trollflow2.plugins import metadata_alias
        with mock.patch('trollflow2.plugins.Pass') as ts_pass,\
                mock.patch('trollflow2.plugins.get_twilight_poly'),\
                mock.patch('trollflow2.plugins.get_area_def'),\
                mock.patch("trollflow2.plugins._get_sunlight_coverage") as _get_sunlight_coverage:
            job = {}
            scene = _get_mocked_scene_with_properties()
            job['scene'] = scene
            job['product_list'] = self.product_list.copy()
            job['input_mda'] = self.input_mda.copy()
            metadata_alias(job)

            job['resampled_scenes'] = {}
            for area in job['product_list']['product_list']['areas']:
                job['resampled_scenes'][area] = {}
            # Run without any settings
            check_sunlight_coverage(job)

            _get_sunlight_coverage.assert_not_called()
            ts_pass.assert_not_called()

    def test_sunlight_filter(self):
        """Test that product isn't loaded when sunlight coverage is to low."""
        from trollflow2.plugins import check_sunlight_coverage
        from trollflow2.plugins import metadata_alias
        with mock.patch('trollflow2.plugins.Pass'),\
                mock.patch('trollflow2.plugins.get_twilight_poly'),\
                mock.patch('trollflow2.plugins.get_area_def'),\
                mock.patch("trollflow2.plugins._get_sunlight_coverage") as _get_sunlight_coverage:
            job = {}
            scene = _get_mocked_scene_with_properties()
            job['scene'] = scene
            job['product_list'] = self.product_list.copy()
            job['input_mda'] = self.input_mda.copy()
            metadata_alias(job)

            job['resampled_scenes'] = {}
            for area in job['product_list']['product_list']['areas']:
                job['resampled_scenes'][area] = {}
            job['product_list']['product_list']['sunlight_coverage'] = {'min': 10, 'max': 40}
            green_snow = mock.MagicMock()
            green_snow.name = 'Green Snow Mock'
            job['resampled_scenes']['euron1']['green_snow'] = green_snow
            green_snow.attrs.__getitem__.return_value = 'euron1'
            # Run without any settings
            _get_sunlight_coverage.return_value = .3
            check_sunlight_coverage(job)

            pl_green = job['product_list']['product_list']['areas']['euron1']['products']['green_snow']

            _get_sunlight_coverage.assert_called_once()
            self.assertIn('green_snow', job['product_list']['product_list']['areas']['euron1']['products'])

            _get_sunlight_coverage.return_value = 0
            with self.assertLogs("trollflow2.plugins", level=logging.INFO) as cm:
                check_sunlight_coverage(job)
            self.assertNotIn('green_snow', job['product_list']['product_list']['areas']['euron1']['products'])
            self.assertIn("Not enough sunlight coverage for product "
                          "'green_snow', removed. Needs at least 10.0%, got "
                          "0.0%.", cm.output[0])

            job['product_list']['product_list']['areas']['euron1']['products']['green_snow'] = pl_green
            _get_sunlight_coverage.return_value = 1
            with self.assertLogs("trollflow2.plugins", level=logging.INFO) as cm:
                check_sunlight_coverage(job)
            self.assertNotIn('green_snow', job['product_list']['product_list']['areas']['euron1']['products'])
            self.assertIn("Too much sunlight coverage for product "
                          "'green_snow', removed. Needs at most 40.0%, got "
                          "100.0%.", cm.output[0])


def _get_mocked_scene_with_properties():
    scene = mock.MagicMock()
    scene.attrs = {}
    scene.start_time = SCENE_START_TIME
    scene.end_time = SCENE_END_TIME
    scene.sensor_names = {'sensor'}

    return scene


class TestCovers(TestCase):
    """Test case for coverage checks."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)
        self.input_mda = {"platform_name": "NOAA-15",
                          "sensor": "avhrr-3",
                          "start_time": dt.datetime(2019, 1, 19, 11),
                          "end_time": dt.datetime(2019, 1, 19, 12),
                          }

    @mock.patch('trollflow2.plugins.Pass', new=None)
    def test_covers_no_trollsched(self):
        """Test coverage when pytroll schedule is missing."""
        from trollflow2.plugins import covers
        job_orig = {"foo": "bar"}
        job = job_orig.copy()
        covers(job)
        self.assertEqual(job, job_orig)

    def test_covers_complains_when_multiple_sensors_are_provided(self):
        """Test that the plugin complains when multiple sensors are provided."""
        from trollflow2.plugins import covers

        with mock.patch('trollflow2.plugins.get_scene_coverage') as get_scene_coverage, \
                mock.patch('trollflow2.plugins.Pass'):
            get_scene_coverage.return_value = 10.0
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": {"platform_name": "platform",
                                 "sensor": ["avhrr-3", "mhs"]},
                   "scene": scn}
            with self.assertLogs("trollflow2.plugins", logging.WARNING) as log:
                covers(job)
            assert len(log.output) == 1
            assert ("Multiple sensors given, taking the first one for coverage calculations" in log.output[0])

    def test_covers_does_not_complain_when_one_sensor_is_provided_as_a_sequence(self):
        """Test that the plugin complains when multiple sensors are provided."""
        from trollflow2.plugins import covers

        with mock.patch('trollflow2.plugins.get_scene_coverage') as get_scene_coverage, \
                mock.patch('trollflow2.plugins.Pass'):
            get_scene_coverage.return_value = 10.0
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": {"platform_name": "platform",
                                 "sensor": ["avhrr-3"]},
                   "scene": scn}
            with self.assertLogs("trollflow2.plugins", logging.WARNING) as log:
                covers(job)
                logger = logging.getLogger("trollflow2.plugins")
                logger.warning("Dummy warning")
            assert len(log.output) == 1

    def test_metadata_is_read_from_scene(self):
        """Test that the scene and message metadata are merged correctly."""
        from trollflow2.plugins import covers

        with mock.patch('trollflow2.plugins.get_scene_coverage') as get_scene_coverage, \
                mock.patch('trollflow2.plugins.Pass'):
            get_scene_coverage.return_value = 10.0
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": {"platform_name": "platform"},
                   "scene": scn}
            covers(job)
            get_scene_coverage.assert_called_with(job["input_mda"]["platform_name"],
                                                  scn.start_time,
                                                  scn.end_time,
                                                  list(scn.sensor_names)[0],
                                                  "omerc_bb")

    def test_covers(self):
        """Test coverage."""
        from trollflow2.plugins import covers
        with mock.patch('trollflow2.plugins.Pass', spec=True) as pass_obj:
            fake_area_coverage_10 = partial(fake_area_coverage, result=.1)
            pass_obj.return_value.area_coverage.side_effect = fake_area_coverage_10
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": self.input_mda,
                   "scene": scn}
            with self.assertLogs("trollflow2.plugins", logging.DEBUG) as log:
                covers(job)
            assert ("DEBUG:trollflow2.plugins:Area coverage 100.00% "
                    "above threshold 5.00% - Carry on with omerc_bb" in log.output)
            # Area "euron1" should be removed
            assert "euron1" not in job['product_list']['product_list']['areas']
            # Other areas should stay in the list
            assert "germ" in job['product_list']['product_list']['areas']
            assert "omerc_bb" in job['product_list']['product_list']['areas']

    def test_covers_uses_only_one_sensor(self):
        """Test that only one sensor is used."""
        from trollflow2.plugins import covers
        input_mda = self.input_mda.copy()
        input_mda['sensor'] = {'avhrr-4'}
        scn = _get_mocked_scene_with_properties()

        job = {"product_list": self.product_list,
               "input_mda": input_mda,
               "scene": scn}
        job2 = copy.deepcopy(job)

        with mock.patch('trollflow2.plugins.get_scene_coverage') as get_scene_coverage,\
                mock.patch('trollflow2.plugins.Pass'):
            get_scene_coverage.return_value = 10.0
            covers(job)
            get_scene_coverage.assert_called_with(input_mda['platform_name'],
                                                  input_mda['start_time'],
                                                  input_mda['end_time'],
                                                  'avhrr-4', 'omerc_bb')

            del job2["product_list"]["product_list"]["areas"]["euron1"]["min_coverage"]
            del job2["product_list"]["product_list"]["min_coverage"]
            with self.assertLogs(level="DEBUG") as log:
                covers(job2)
                assert "Minimum area coverage not given" in log.output[0]

    def test_scene_coverage(self):
        """Test scene coverage."""
        from trollflow2.plugins import get_scene_coverage
        with mock.patch('trollflow2.plugins.get_area_def') as get_area_def,\
                mock.patch('trollflow2.plugins.Pass') as ts_pass:
            area_coverage = mock.MagicMock()
            area_coverage.return_value = 0.2
            overpass = mock.MagicMock()
            overpass.area_coverage = area_coverage
            ts_pass.return_value = overpass
            get_area_def.return_value = 6
            res = get_scene_coverage(1, 2, 3, 4, 5)
            self.assertEqual(res, 100 * 0.2)
            ts_pass.assert_called_with(1, 2, 3, instrument=4)
            get_area_def.assert_called_with(5)
            area_coverage.assert_called_with(6)

    def test_covers_collection_area_id(self):
        """Test the coverage of a collection area id."""
        from trollflow2.plugins import covers
        from trollflow2.plugins import AbortProcessing
        with mock.patch('trollflow2.plugins.Pass', spec=True) as pass_obj:
            fake_area_coverage_100 = partial(fake_area_coverage, result=1)
            pass_obj.return_value.area_coverage.side_effect = fake_area_coverage_100
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": self.input_mda,
                   "scene": scn}
            # Nothing should happen here
            covers(job)
            # Area that matches the product list, nothing should happen
            job['input_mda']['collection_area_id'] = 'euron1'
            covers(job)
            # By default collection_area_id isn't checked so nothing should happen
            job['input_mda']['collection_area_id'] = 'not_in_pl'
            covers(job)
            # Turn coverage check on, so area not in the product list
            # should raise AbortProcessing
            job['product_list']['product_list']['coverage_by_collection_area'] = True
            with self.assertRaises(AbortProcessing):
                covers(job)

            # And with existing area there shouldn't be an exception
            job['input_mda']['collection_area_id'] = 'euron1'
            covers(job)

    def test_covers_returns_100_when_area_def_is_dynamic(self):
        """Test that covers return 100% when area def is dynamic."""
        from trollflow2.plugins import covers
        with mock.patch('trollflow2.plugins.Pass', spec=True) as pass_obj:
            pass_obj.return_value.area_coverage.side_effect = partial(fake_area_coverage, result=.1)
            scn = _get_mocked_scene_with_properties()
            job = {"product_list": self.product_list,
                   "input_mda": self.input_mda,
                   "scene": scn}
            covers(job)
            assert job["product_list"]["product_list"]["areas"]["omerc_bb"]["area_coverage_percent"] == 100


def fake_area_coverage(areadef, result=.1):
    """Fake area coverage."""
    if isinstance(areadef, DynamicAreaDefinition):
        raise AttributeError
    else:
        return result


class TestCheckMetadata(TestCase):
    """Test case for checking the input metadata."""

    def test_single_item(self):
        """Test checking a single metadata item."""
        from trollflow2.plugins import check_metadata
        from trollflow2.plugins import AbortProcessing
        with mock.patch('trollflow2.plugins.get_config_value') as get_config_value:
            get_config_value.return_value = None
            job = {'product_list': None, 'input_mda': {'sensor': 'foo'}}
            self.assertIsNone(check_metadata(job))
            get_config_value.return_value = {'sensor': ['foo', 'bar']}
            self.assertIsNone(check_metadata(job))
            get_config_value.return_value = {'sensor': ['bar']}
            with self.assertRaises(AbortProcessing):
                check_metadata(job)

    def test_multiple_items(self):
        """Test checking a single metadata item."""
        from trollflow2.plugins import check_metadata
        from trollflow2.plugins import AbortProcessing
        with mock.patch('trollflow2.plugins.get_config_value') as get_config_value:
            # Nothing configured
            get_config_value.return_value = None
            job = {'product_list': None,
                   'input_mda': {'sensor': 'foo',
                                 'platform_name': 'bar'}}
            self.assertIsNone(check_metadata(job))
            # Both sensor and platform name match
            get_config_value.return_value = {'sensor': ['foo', 'bar'],
                                             'platform_name': ['bar']}
            self.assertIsNone(check_metadata(job))
            # Sensor matches, 'variant' not in the message
            get_config_value.return_value = {'sensor': ['foo', 'bar'],
                                             'variant': ['e ascari']}
            self.assertIsNone(check_metadata(job))
            # Platform doesn't match -> abort
            get_config_value.return_value = {'sensor': ['foo'],
                                             'platform_name': ['not-bar']}
            with self.assertRaises(AbortProcessing):
                check_metadata(job)


class TestMetadataAlias(TestCase):
    """Test case for metadata alias."""

    def test_metadata_alias(self):
        """Test metadata aliasing."""
        from trollflow2.plugins import metadata_alias
        mda = {'platform_name': 'noaa15', 'not_changed': True}
        product_list = {'common': {'not_metadata_aliases': True}}
        job = {'input_mda': mda, 'product_list': product_list}
        metadata_alias(job)
        mda = job['input_mda']
        self.assertEqual(mda['platform_name'], 'noaa15')
        self.assertTrue(mda['not_changed'])
        product_list = {'common': {'metadata_aliases':
                                   {'platform_name': {'noaa15': 'NOAA-15'},
                                    'not_in_mda': {'something': 'other'}}}}
        job = {'input_mda': mda, 'product_list': product_list}
        metadata_alias(job)
        mda = job['input_mda']
        self.assertEqual(mda['platform_name'], 'NOAA-15')
        self.assertTrue(mda['not_changed'])
        self.assertTrue('not_in_mda' not in mda)

    def test_iterable_metadata(self):
        """Test that iterable metadata gets replaced."""
        from trollflow2.plugins import metadata_alias
        mda = {'sensor': ('a/b',), 'foo': set(['c/d'])}
        product_list = {'common': {'metadata_aliases':
                                   {'sensor': {'a/b': 'a-b'},
                                    'foo': {'c/d': 'c-d'}}}}
        job = {'input_mda': mda, 'product_list': product_list}
        metadata_alias(job)
        self.assertEqual(job['input_mda']['sensor'], ('a-b',))
        self.assertEqual(job['input_mda']['foo'], set(['c-d']))


class TestGetPluginConf(TestCase):
    """Test case for get_plugin_conf."""

    def test_get_plugin_conf(self):
        """Test the get_plugin_conf function."""
        from trollflow2.plugins import _get_plugin_conf
        conf = {"common": {"val1": "foo1"},
                "product_list": {"val2": "bar2"}}
        path = "/product_list"
        defaults = {"val1": "foo0", "val2": "bar0", "val3": "baz0"}
        res = _get_plugin_conf(conf, path, defaults)
        self.assertTrue("val1" in res)
        self.assertTrue("val2" in res)
        self.assertTrue("val3" in res)
        self.assertEqual(res["val1"], "foo1")
        self.assertEqual(res["val2"], "bar2")
        self.assertEqual(res["val3"], "baz0")


class TestSZACheck(TestCase):
    """Test case for SZA check."""

    def setUp(self):
        """Create common items."""
        product_list_no_sza, job_no_sza = _get_product_list_and_job()
        self.product_list_no_sza = product_list_no_sza
        self.job_no_sza = job_no_sza
        product_list_with_sza, job_with_sza = _get_product_list_and_job(add_sza_limits=True)
        self.product_list_with_sza = product_list_with_sza
        self.job_with_sza = job_with_sza

    def test_sza_check_no_settings(self):
        """Test the SZA check without any settings."""
        from trollflow2.plugins import sza_check
        with mock.patch("trollflow2.plugins.sun_zenith_angle") as sun_zenith_angle:
            sza_check(self.job_no_sza)
            sun_zenith_angle.assert_not_called()

    def test_metadata_is_read_from_scene(self):
        """Test that the scene and message metadata are merged correctly."""
        from trollflow2.plugins import sza_check

        with mock.patch("trollflow2.plugins.sun_zenith_angle") as sun_zenith_angle:
            sun_zenith_angle.return_value = 90.
            scn = _get_mocked_scene_with_properties()
            job = self.job_with_sza.copy()
            del job["input_mda"]["start_time"]
            job["scene"] = scn
            sza_check(job)
            sun_zenith_angle.assert_called_with(scn.start_time, 25., 60.)

    def test_sza_check_with_ok_sza(self):
        """Test the SZA check with SZA that is ok for all the products."""
        from trollflow2.plugins import sza_check
        with mock.patch("trollflow2.plugins.sun_zenith_angle") as sun_zenith_angle:
            # Zenith angle that is ok for all the products
            sun_zenith_angle.return_value = 90.

            sza_check(self.job_with_sza)

            sun_zenith_angle.assert_called_with(JOB_INPUT_MDA_START_TIME, 25., 60.)
            self.assertDictEqual(self.job_with_sza['product_list'], self.product_list_with_sza)

    def test_sza_check_removes_day_products(self):
        """Test the SZA check with SZA that removes day products."""
        from trollflow2.plugins import sza_check
        with mock.patch("trollflow2.plugins.sun_zenith_angle") as sun_zenith_angle:
            # Zenith angle that removes day products
            sun_zenith_angle.return_value = 100.

            sza_check(self.job_with_sza)

            self.assertTrue('cloud_top_height' in
                            self.product_list_with_sza['product_list']['areas']['omerc_bb']['products'])
            self.assertFalse('ct' in self.product_list_with_sza['product_list']['areas']['omerc_bb']['products'])

    def test_sza_check_removes_night_products(self):
        """Test the SZA check with SZA that removes night products."""
        from trollflow2.plugins import sza_check
        with mock.patch("trollflow2.plugins.sun_zenith_angle") as sun_zenith_angle:
            # Zenith angle that removes night products
            sun_zenith_angle.return_value = 45.

            sza_check(self.job_with_sza)

            # There was only one product, so the whole area is deleted
            self.assertFalse('germ' in self.job_with_sza['product_list']['product_list']['areas'])


def _get_product_list_and_job(add_sza_limits=False):
    from yaml import UnsafeLoader
    product_list = read_config(raw_string=yaml_test1, Loader=UnsafeLoader)
    if add_sza_limits:
        _add_sunzen_limits(product_list)
    job = _create_job(product_list)

    return product_list, job


def _create_job(product_list):
    job = {}
    scene = _get_mocked_scene_with_properties()
    job['input_mda'] = {'start_time': JOB_INPUT_MDA_START_TIME, 'another_message_item': 'coconut'}
    job['scene'] = scene
    job['product_list'] = product_list.copy()

    return job


def _add_sunzen_limits(product_list):
    # Add SZA limits to couple of products
    # Day product
    product_list['product_list']['areas']['omerc_bb']['products']['ct']['sunzen_maximum_angle'] = 95.
    product_list['product_list']['areas']['omerc_bb']['products']['ct']['sunzen_check_lon'] = 25.
    product_list['product_list']['areas']['omerc_bb']['products']['ct']['sunzen_check_lat'] = 60.
    # Night product
    product_list['product_list']['areas']['germ']['products']['cloudtype']['sunzen_minimum_angle'] = 85.
    product_list['product_list']['areas']['germ']['products']['cloudtype']['sunzen_check_lon'] = 25.
    product_list['product_list']['areas']['germ']['products']['cloudtype']['sunzen_check_lat'] = 60.


class TestOverviews(TestCase):
    """Test case for overviews."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import BaseLoader
        self.product_list = read_config(raw_string=yaml_test1, Loader=BaseLoader)

    def test_add_overviews(self):
        """Test adding overviews."""
        from trollflow2.plugins import add_overviews
        with mock.patch('trollflow2.plugins.Resampling') as resampling,\
                mock.patch('trollflow2.plugins.rasterio') as rasterio:
            # Mock the rasterio.open context manager
            dst = mock.MagicMock()
            rasterio.open.return_value.__enter__.return_value = dst

            product_list = self.product_list['product_list']['areas']
            product_list['germ']['products']['cloudtype']['formats'][0]['overviews'] = [4]
            # Add filename, otherwise added by `save_datasets()`
            product_list['germ']['products']['cloudtype']['formats'][0]['filename'] = 'foo'
            job = {"product_list": self.product_list}
            add_overviews(job)
            dst.build_overviews.assert_called_once_with([4], resampling.average)
            dst.update_tags.assert_called_once_with(ns='rio_overview',
                                                    resampling='average')


class TestFilePublisher(TestCase):
    """Test case for File publisher."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        from yaml import UnsafeLoader
        self.product_list = read_config(raw_string=yaml_test_publish, Loader=UnsafeLoader)
        # Skip omerc_bb area, there's no fname_pattern
        del self.product_list['product_list']['areas']['omerc_bb']
        self.input_mda = input_mda.copy()
        self.input_mda['uri'] = 'foo.nc'

    def test_filepublisher_is_started(self):
        """Test that the filepublisher is started."""
        from trollflow2.plugins import FilePublisher
        with mock.patch('trollflow2.plugins.NoisyPublisher'):
            pub = FilePublisher()
            pub.pub.start.assert_called_once()

    def test_filepublisher_is_stopped_on_exit(self):
        """Test that the filepublisher is stopped on exit."""
        from trollflow2.plugins import FilePublisher
        with mock.patch('trollflow2.plugins.NoisyPublisher'):
            pub = FilePublisher()
            pub.__del__()
            pub.pub.stop.assert_called()

    def test_filepublisher_with_compose(self):
        """Test filepublisher with compose."""
        from trollflow2.plugins import FilePublisher
        from satpy import Scene
        from satpy.tests.utils import make_dataid

        scn_euron1 = Scene()
        dataid = make_dataid(name='cloud_top_height', resolution=1000)
        scn_euron1[dataid] = mock.MagicMock()
        job = {'product_list': self.product_list,
               'input_mda': self.input_mda,
               'resampled_scenes': dict(euron1=scn_euron1)}

        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            message = mocks['Message']

            pub = FilePublisher()
            product_list = self.product_list.copy()
            product_list['product_list']['publish_topic'] = '/{areaname}/{productname}'
            topics = self._create_filenames_and_topics(job)

            pub(job)
            message.assert_called()
            pub.pub.send.assert_called()

            call_count = 0
            for area in job['product_list']['product_list']['areas']:
                for _prod in job['product_list']['product_list']['areas'][area]:
                    # Skip calls to __str__
                    if 'call().__str__()' != str(message.mock_calls[call_count]):
                        self.assertTrue(topics[call_count] in str(message.mock_calls[call_count]))
                        call_count += 1
            self.assertEqual(call_count, 1)
            self.assertEqual(message.call_args[0][2]['processing_center'], 'SMHI')

    def test_filepublisher_without_compose(self):
        """Test filepublisher without compose."""
        from satpy import Scene
        from satpy.tests.utils import make_dataid

        scn_euron1 = Scene()
        dataid = make_dataid(name='cloud_top_height', resolution=1000)
        scn_euron1[dataid] = mock.MagicMock()
        job = {'product_list': self.product_list,
               'input_mda': self.input_mda,
               'resampled_scenes': dict(euron1=scn_euron1)}

        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            message = mocks['Message']

            pub, topics = self._run_publisher_on_job(job)
            message.assert_called()
            pub.pub.send.assert_called()

            call_count = 0
            for area in job['product_list']['product_list']['areas']:
                for _prod in job['product_list']['product_list']['areas'][area]:
                    # Skip calls to __str__
                    if 'call().__str__()' != str(message.mock_calls[call_count]):
                        self.assertTrue(topics[call_count] in str(message.mock_calls[call_count]))
                        call_count += 1
            self.assertEqual(call_count, 1)

    def test_non_existing_products_are_not_published(self):
        """Test that non existing products are not published."""
        from satpy import Scene

        scn = _get_mocked_scene_with_properties()
        job = {"scene": scn, "product_list": self.product_list, 'input_mda': self.input_mda,
               'resampled_scenes': dict(euron1=Scene(), germ=Scene())}

        with mock.patch('trollflow2.plugins.Message') as message, mock.patch('trollflow2.plugins.NoisyPublisher'):
            self._run_publisher_on_job(job)
            message.assert_not_called()

    def test_multiple_dataset_files_can_be_published(self):
        """Test that netcdf files with multiple datasets can be published normally."""
        from satpy import Scene
        import numpy as np
        from satpy.tests.utils import make_dataid

        resampled_scene = Scene()
        resampled_scene[make_dataid(name='latitude')] = np.ones((4, 4))

        scn = _get_mocked_scene_with_properties()
        job = {"scene": scn, "product_list": self.product_list, 'input_mda': self.input_mda,
               'resampled_scenes': {'None': resampled_scene}}

        with mock.patch('trollflow2.plugins.Message') as message, mock.patch('trollflow2.plugins.NoisyPublisher'):
            self._run_publisher_on_job(job)
            assert message.call_args_list[-1][0][2]['product'] == (
                'chl_nn', 'chl_oc4me', 'trsp', 'tsm_nn', 'iop_nn', 'mask', 'latitude', 'longitude')

    def _run_publisher_on_job(self, job):
        """Run a publisher on *job*."""
        from trollflow2.plugins import FilePublisher

        pub = FilePublisher()
        product_list = self.product_list.copy()
        product_list['product_list']['publish_topic'] = '/static_topic'
        topics = self._create_filenames_and_topics(job)
        pub(job)
        return pub, topics

    @staticmethod
    def _create_filenames_and_topics(job):
        """Create the filenames and topics for *job*."""
        from trollflow2.dict_tools import plist_iter
        from trollsift import compose
        import os.path

        topic_pattern = job['product_list']['product_list']['publish_topic']
        topics = []

        for fmat, fmat_config in plist_iter(job['product_list']['product_list'],
                                            job['input_mda'].copy()):
            fname_pattern = fmat['fname_pattern']
            filename = compose(os.path.join(fmat['output_dir'],
                                            fname_pattern), fmat)
            fmat.pop('format', None)
            fmat_config['filename'] = filename
            topics.append(compose(topic_pattern, fmat))

        return topics

    def test_filepublisher_kwargs(self):
        """Test filepublisher keyword argument usage."""
        from yaml import UnsafeLoader
        from trollflow2.plugins import FilePublisher

        # Direct instantiation
        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            NoisyPublisher = mocks['NoisyPublisher']
            Publisher = mocks['Publisher']

            pub = FilePublisher()
            pub.pub.start.assert_called_once()
            assert mock.call('l2processor', port=0, nameservers="") in NoisyPublisher.mock_calls
            Publisher.assert_not_called()
            assert pub.port == 0
            assert pub.nameservers == ""
            pub = FilePublisher(port=40000, nameservers=['localhost'])
            assert mock.call('l2processor', port=40000,
                             nameservers=['localhost']) in NoisyPublisher.mock_calls
            assert pub.port == 40000
            assert pub.nameservers == ['localhost']
            assert len(pub.pub.start.mock_calls) == 2

        # Direct instantiation with nameservers set to None, which should use Publisher instead of NoisyPublisher
        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            NoisyPublisher = mocks['NoisyPublisher']
            Publisher = mocks['Publisher']

            pub = FilePublisher(port=40000, nameservers=None)
            NoisyPublisher.assert_not_called()
            Publisher.assert_called_once_with('tcp://*:40000', 'l2processor')

        # Instantiate via loading YAML
        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            NoisyPublisher = mocks['NoisyPublisher']
            Publisher = mocks['Publisher']

            fpub = read_config(raw_string=YAML_FILE_PUBLISHER, Loader=UnsafeLoader)
            assert mock.call('l2processor', port=40002,
                             nameservers=['localhost']) in NoisyPublisher.mock_calls
            Publisher.assert_not_called()
            fpub.pub.start.assert_called_once()
            assert fpub.port == 40002
            assert fpub.nameservers == ['localhost']

    def test_dispatch(self):
        """Test dispatch order messages."""
        from trollflow2.plugins import FilePublisher
        from satpy import Scene
        from satpy.tests.utils import make_dataid

        scn = Scene()
        dataid = make_dataid(name='cloud_top_height', resolution=1000)
        scn[dataid] = mock.MagicMock()
        job = {'product_list': self.product_list,
               'input_mda': self.input_mda,
               'resampled_scenes': dict(euron1=scn)}

        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            message = mocks['Message']

            pub = FilePublisher()
            pub(job)
            dispatches = 0
            for args, _kwargs in message.call_args_list:
                mda = args[2]
                if args[1] == 'file':
                    self.assertIn('uri', mda)
                    self.assertIn('uid', mda)
                elif args[1] == 'dispatch':
                    self.assertIn('source', mda)
                    self.assertIn('target', mda)
                    self.assertIn('file_mda', mda)
                    self.assertEqual(mda['source'],
                                    '/tmp/satdmz/pps/www/latest_2018/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png')  # noqa
                    self.assertEqual(mda['target'],
                                    'ftp://ftp.important_client.com/somewhere/NOAA-15_20190217_0600_euron1_in_fname_ctth_static.png')  # noqa
                    dispatches += 1
            self.assertEqual(dispatches, 1)

    def test_deleting(self):
        """Test deleting the publisher."""
        from trollflow2.plugins import FilePublisher
        nb_ = mock.MagicMock()
        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            NoisyPublisher = mocks['NoisyPublisher']

            NoisyPublisher.return_value = nb_
            pub = FilePublisher()
            job = {'product_list': self.product_list,
                   'input_mda': self.input_mda,
                   'resampled_scenes': {}}
            pub(job)

        nb_.stop.assert_not_called()
        del pub
        nb_.stop.assert_called_once()

    def test_stopping(self):
        """Test stopping the publisher."""
        from trollflow2.plugins import FilePublisher
        nb_ = mock.MagicMock()
        with mock.patch.multiple('trollflow2.plugins', Message=mock.DEFAULT,
                                 NoisyPublisher=mock.DEFAULT, Publisher=mock.DEFAULT) as mocks:
            NoisyPublisher = mocks['NoisyPublisher']

            NoisyPublisher.return_value = nb_
            pub = FilePublisher()
            job = {'product_list': self.product_list,
                   'input_mda': self.input_mda,
                   'resampled_scenes': {}}
            pub(job)

        nb_.stop.assert_not_called()
        pub.stop()
        nb_.stop.assert_called_once()


class FakeScene(dict):
    """Scene drop-in replacement, just a dict that can have attributes."""


@pytest.fixture
def sc_3a_3b():
    """Fixture to prepare a scene with channels 3A and 3B."""
    from xarray import DataArray
    from satpy import Scene
    import dask.array as da
    import numpy as np
    prod_attrs = {
        "platform_name": "noaa-18",
        "sensor": "avhrr-3"}
    scene = Scene()
    # NB: Scene.__setattr__ will turn this into a DataID.  This means that
    # after ``scene["NIR016"] = x``, we still have "NIR016" not in
    # scene.keys() (but "NIR016" in scene).
    scene["NIR016"] = DataArray(
        da.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan], [0.5, 0.5, 0.5]]),
        dims=("y", "x"),
        attrs=prod_attrs)
    scene["IR037"] = DataArray(
        np.array([[200, 230, 240], [250, 260, 220], [np.nan, np.nan, np.nan]]),
        dims=("y", "x"),
        attrs=prod_attrs)
    scene["another"] = DataArray(
        da.array([[200, 230, 240], [250, 260, 220], [1, 2, 3]]),
        dims=("y", "x"),
        attrs=prod_attrs)
    scene["NIR016"].attrs.update(
        start_time=dt.datetime(2019, 1, 19, 13),
        end_time=dt.datetime(2019, 1, 19, 13))
    scene["IR037"].attrs.update(
        start_time=dt.datetime(2019, 1, 19, 11),
        end_time=dt.datetime(2019, 1, 19, 12))
    scene.attrs = {}
    return scene


def test_valid_filter(caplog, sc_3a_3b):
    """Test filter for minimum fraction of valid data."""
    from trollflow2.launcher import yaml
    from trollflow2.plugins import check_valid_data_fraction
    product_list = yaml.safe_load(yaml_test3)

    job = {}
    job['scene'] = sc_3a_3b
    job['product_list'] = product_list.copy()
    job['input_mda'] = input_mda.copy()
    job['resampled_scenes'] = {"euron1": sc_3a_3b}
    prods = job['product_list']['product_list']['areas']['euron1']['products']
    for p in ("NIR016", "IR037", "absent"):
        prods[p] = {"min_valid_data_fraction": 40}
    job2 = copy.deepcopy(job)
    prods2 = job2['product_list']['product_list']['areas']['euron1']['products']

    with mock.patch("trollflow2.plugins.get_scene_coverage") as tpg, \
            caplog.at_level(logging.DEBUG):
        tpg.return_value = 100
        check_valid_data_fraction(job)
        assert "NIR016" not in prods
        assert "IR037" in prods
        assert "removing NIR016 for area euron1" in caplog.text
        assert "keeping IR037 for area euron1" in caplog.text
        assert "product absent not found, already removed" in caplog.text
        tpg.reset_mock()
        tpg.return_value = 1
        check_valid_data_fraction(job2)
        assert "inaccurate coverage estimate suspected!" in caplog.text
        assert "NIR016" in prods2
        assert "IR037" in prods2
        tpg.reset_mock()
        tpg.return_value = 0
        check_valid_data_fraction(job2)
        assert "no expected coverage at all, removing" in caplog.text
        assert "NIR016" not in prods2
        assert "IR037" not in prods2


def test_persisted(sc_3a_3b):
    """Test that early persisting does what we want."""
    from trollflow2.launcher import yaml
    from trollflow2.plugins import _persist_what_we_must
    job = {}
    product_list = yaml.safe_load(yaml_test3)
    job['product_list'] = product_list.copy()
    job['input_mda'] = input_mda.copy()
    job['resampled_scenes'] = {"euron1": sc_3a_3b}
    prods = job['product_list']['product_list']['areas']['euron1']['products']
    for p in ("NIR016", "IR037", "absent"):
        prods[p] = {"min_valid_data_fraction": 40}

    def fake_persist(*args):
        for da in args:
            da.attrs["persisted"] = True
        return args

    with mock.patch("dask.persist", new=fake_persist):
        _persist_what_we_must(job)

    # confirm that sc_3a_3b dataset NIR016 is persisted
    assert sc_3a_3b["NIR016"].attrs.get("persisted")
    assert not sc_3a_3b["another"].attrs.get("persisted")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Pytroll developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

# Workaround for unittests so that satpy and posttroll installations
# are not necessary
"""Tools for product list operations."""

import dpath.util


def plist_iter(product_list, base_mda=None, level=None):
    """Iterate over the product list at a configurable level.

    This function walks through the product-list (depth-wise) and yields
    two configuration items each time: a flattened version of the configurations
    accumulated accross the levels and the current item's configuration.

    *base_mda* provides the base configuration to include, and *level* (one of
    'area', 'product', or None (default, all levels included)) is the max depth
    to walk the product list at.
    """
    if base_mda is None:
        base_mda = {}
    else:
        base_mda = base_mda.copy()
    for area, area_config in product_list['areas'].items():
        aconfig = base_mda.copy()
        aconfig.update(product_list)
        aconfig.pop('areas', None)
        aconfig.update(area_config)
        aconfig.pop('products', None)
        aconfig['area'] = area
        if level == 'area':
            yield aconfig, area_config
            continue
        for prod, prod_config in area_config['products'].items():
            pconfig = aconfig.copy()
            pconfig.update(prod_config)
            pconfig['product'] = prod
            if level == 'product':
                yield pconfig, prod_config
                continue
            for file_config in pconfig.get('formats', [{'format': 'tif', 'writer': 'geotiff'}]):
                fconfig = pconfig.copy()
                fconfig.pop('formats', None)
                fconfig.update(file_config)
                yield fconfig, file_config


def gen_dict_extract(var, key):
    """Generate the values of *key* recusively from the dict *var*."""
    if hasattr(var, 'items'):
        for k, v in var.items():
            if k == key:
                yield v
            if hasattr(v, 'items'):
                for result in gen_dict_extract(v, key):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in gen_dict_extract(d, key):
                        yield result


def get_config_value(config, path, key, default=None):
    """Get the most local config value for key *key* starting from the dictionary path *path*.

    If nothing is found, path "/common/" is also checked, and if still nothing is found, return *default*.
    """
    path_parts = path.split('/')
    # Loop starting from the current path, and continue upwards
    # towards the root until something is found
    num = len(path_parts)
    for i in range(num, 1, -1):
        pwd = "/".join(path_parts[:i] + [key])
        vals = dpath.util.values(config, pwd)
        if len(vals) > 0:
            return vals[0]

    vals = dpath.util.values(config, "/common/" + key)
    if len(vals) > 0:
        return vals[0]

    return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pytroll developers
#
# Author(s):
#
#   Martin Raspaud <martin.raspaud@smhi.se>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
"""Tests for the logging utilities."""

import logging
import sys
import time
from multiprocessing import Manager
from unittest import mock

import pytest
from trollflow2.logging import logging_on, setup_queued_logging

log_queue = Manager().Queue(-1)  # no limit on size


def test_queued_logging_has_a_listener():
    """Test that the queued logging has a listener."""
    with mock.patch("trollflow2.logging.QueueListener", autospec=True) as q_listener:
        with logging_on(log_queue):
            assert q_listener.called
            assert q_listener.return_value.start.called
        assert q_listener.return_value.stop.called


def test_queued_logging_stops_listener_on_exception():
    """Test that queued logging stops the listener even if an exception occurs."""
    with mock.patch("trollflow2.logging.QueueListener", autospec=True) as q_listener:
        with pytest.raises(Exception):
            with logging_on(log_queue):
                raise Exception("Oh no!")
        assert q_listener.return_value.stop.called


LOG_CONFIG = {'version': 1,
              'formatters': {'simple': {'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'}},
              'handlers': {'file': {'class': 'logging.handlers.BufferingHandler',
                                    'capacity': 1,
                                    'formatter': 'simple'}},
              'root': {'level': 'INFO', 'handlers': ['file']}}


def test_log_config_is_used_when_provided():
    """Test that the log config is used when provided."""
    config = LOG_CONFIG

    log = logging.getLogger()
    with mock.patch("logging.handlers.BufferingHandler.emit", autospec=True) as emit:
        with logging_on(log_queue, config=config):
            assert not emit.called
            log.warning("uh oh...")
            # we wait for the log record to go through the queue listener in
            # its own thread
            time.sleep(.01)
            assert emit.called


def test_logging_works(caplog):
    """Test that the logs get out there."""
    log = logging.getLogger("something")
    with logging_on(log_queue):
        logging.getLogger().addHandler(caplog.handler)
        log.warning("oh no :(")
        assert "oh no :(" in caplog.text


def fun(q, log_message):
    """Fake a function to run."""
    log = logging.getLogger('for fun')
    setup_queued_logging(q)
    log.debug(log_message)


def run_subprocess(log_message, queue):
    """Run a subprocess."""
    from multiprocessing import get_context
    ctx = get_context('spawn')
    proc = ctx.Process(target=fun, args=(queue, log_message))
    proc.start()
    proc.join()


@pytest.mark.skipif(sys.platform != "linux",
                    reason="Logging from a subprocess seems to work only on Linux")
def test_logging_works_in_subprocess(caplog):
    """Test that the logs get out there, even from a subprocess."""
    log_message = 'yeah, we are in a subprocess now'
    with logging_on(log_queue):
        logging.getLogger().addHandler(caplog.handler)

        run_subprocess(log_message, log_queue)
        assert log_message in caplog.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Pytroll developers
#
# Author(s):
#
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
"""Test the launcher module."""

import datetime
import logging
import queue
import sys
import time
import unittest
from contextlib import contextmanager
from logging.handlers import QueueHandler
import multiprocessing

import pytest
import yaml
from yaml import YAMLError

try:
    from yaml import UnsafeLoader
except ImportError:
    from yaml import Loader as UnsafeLoader
from unittest import mock
from trollflow2.tests.utils import TestCase
from trollflow2.launcher import process

yaml_test1 = """
product_list:
  something: foo
  min_coverage: 5.0
  subscribe_topics:
    - /topic1
    - /topic2
  areas:
      euron1:
        areaname: euron1
        min_coverage: 20.0
        priority: 1
        products:
          ctth:
            productname: cloud_top_height
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image
              - format: jpg
                writer: simple_image
                fill_value: 0
            fname_pattern: "{platform_name:s}_{start_time:%Y%m%d_%H%M}_{areaname:s}_ctth.{format}"

      germ:
        areaname: germ
        fname_pattern: "{start_time:%Y%m%d_%H%M}_{areaname:s}_{productname}.{format}"
        products:
          cloudtype:
            productname: cloudtype
            output_dir: /tmp/satdmz/pps/www/latest_2018/
            formats:
              - format: png
                writer: simple_image

      omerc_bb:
        areaname: omerc_bb
        output_dir: /tmp
        products:
          ct:
            productname: ct
            formats:
              - format: nc
                writer: cf
          cloud_top_height:
            productname: cloud_top_height
            formats:
              - format: tif
                writer: geotiff
"""

yaml_test_minimal = """
product_list:
  output_dir: &output_dir
    /mnt/output/
  publish_topic: /MSG_0deg/L3
  reader: seviri_l1b_hrit
  fname_pattern:
    "{start_time:%Y%m%d_%H%M}_{platform_name}_{areaname}_{productname}.{format}"
  formats:
    - format: tif
      writer: geotiff
  areas:
      euro4:
        areaname: euro4
        products:
          overview:
            productname: overview
          airmass:
            productname: airmass
          natural_color:
            productname: natural_color
          night_fog:
            productname: night_fog

workers:
  - fun: !!python/name:trollflow2.plugins.create_scene
  - fun: !!python/name:trollflow2.plugins.load_composites
  - fun: !!python/name:trollflow2.plugins.resample
  - fun: !!python/name:trollflow2.plugins.save_datasets
#  - fun: !!python/object:trollflow2.FilePublisher {}
"""


class TestGetAreaPriorities(TestCase):
    """Test case for area priorities."""

    def test_get_area_priorities(self):
        """Test getting the area priorities."""
        from trollflow2.launcher import get_area_priorities
        prodlist = yaml.load(yaml_test1, Loader=UnsafeLoader)

        priorities = get_area_priorities(prodlist)
        self.assertTrue(1 in priorities)
        self.assertTrue(isinstance(priorities[1], list))
        self.assertTrue('euron1' in priorities[1])
        self.assertTrue(999 in priorities)
        self.assertTrue(isinstance(priorities[999], list))
        self.assertTrue('omerc_bb' in priorities[999])
        self.assertTrue('germ' in priorities[999])


class TestMessageToJobs(TestCase):
    """Test case for converting a message to jobs."""

    def test_message_to_jobs(self):
        """Test converting message to jobs."""
        from trollflow2.launcher import message_to_jobs
        prodlist = yaml.load(yaml_test1, Loader=UnsafeLoader)
        msg = mock.MagicMock()
        msg.data = {'uri': 'foo'}

        jobs = message_to_jobs(msg, prodlist)
        self.assertEqual(set(jobs.keys()), {1, 999})
        for i in jobs.keys():
            self.assertEqual(set(jobs[i].keys()),
                             {'input_filenames', 'input_mda', 'product_list'})
            self.assertEqual(jobs[i]['input_filenames'], ['foo'])
            self.assertEqual(jobs[i]['input_mda'], msg.data)
            self.assertEqual(set(jobs[i]['product_list'].keys()),
                             {'product_list'})
        self.assertEqual(set(jobs[1]['product_list']['product_list']['areas'].keys()),
                         set(['euron1']))
        self.assertEqual(set(jobs[999]['product_list']['product_list']['areas'].keys()),
                         set(['germ', 'omerc_bb']))

        prodlist['product_list']['areas']['germ']['priority'] = None
        jobs = message_to_jobs(msg, prodlist)
        self.assertTrue('germ' in jobs[999]['product_list']['product_list']['areas'])

    def test_message_to_jobs_minimal(self):
        """Test converting a message to minimal jobs."""
        from trollflow2.launcher import message_to_jobs
        prodlist = yaml.load(yaml_test_minimal, Loader=UnsafeLoader)
        msg = mock.MagicMock()
        msg.data = {'uri': 'foo'}
        jobs = message_to_jobs(msg, prodlist)

        expected = dict([('euro4',
                          {'areaname': 'euro4',
                           'products': {'airmass': {'formats': [{'format': 'tif',
                                                                 'writer': 'geotiff'}],
                                                    'productname': 'airmass'},
                                        'natural_color': {'formats': [{'format': 'tif',
                                                                       'writer': 'geotiff'}],
                                                          'productname': 'natural_color'},
                                        'night_fog': {'formats': [{'format': 'tif',
                                                                   'writer': 'geotiff'}],
                                                      'productname': 'night_fog'},
                                        'overview': {'formats': [{'format': 'tif',
                                                                  'writer': 'geotiff'}],
                                                     'productname': 'overview'}}})])
        self.assertDictEqual(jobs[999]['product_list']['product_list']['areas'], expected)
        self.assertIn('output_dir', jobs[999]['product_list']['product_list'])
        # Test that the formats are not the same object
        prods = jobs[999]['product_list']['product_list']['areas']['euro4']['products']
        self.assertFalse(prods['overview']['formats'][0] is
                         prods['natural_color']['formats'][0])
        prods['overview']['formats'][0]['foo'] = 'bar'
        self.assertFalse('foo' in prods['natural_color']['formats'][0])

    def test_message_to_jobs_fsspec(self):
        """Test transforming a message containing filesystem specification."""
        with mock.patch.dict('sys.modules', {'fsspec': mock.MagicMock(),
                                             'fsspec.spec': mock.MagicMock(),
                                             'satpy': mock.MagicMock(),
                                             'satpy.readers': mock.MagicMock(),
                                             'satpy.resample': mock.MagicMock(),
                                             'satpy.writers': mock.MagicMock(),
                                             'satpy.dataset': mock.MagicMock(),
                                             'satpy.version': mock.MagicMock()}):
            from fsspec.spec import AbstractFileSystem as abs_fs
            from satpy.readers import FSFile as fsfile
            from trollflow2.launcher import message_to_jobs
            import json

            filename = "/S3A_OL_2_WFR____20201210T080758_20201210T080936_20201210T103707_0097_066_078_1980_MAR_O_NR_002.SEN3/Oa01_reflectance.nc"  # noqa
            fs = {"cls": "fsspec.implementations.zip.ZipFileSystem",
                  "protocol": "abstract",
                  "args": ["sentinel-s3-ol2wfr-zips/2020/12/10/S3A_OL_2_WFR____20201210T080758_20201210T080936_20201210T103707_0097_066_078_1980_MAR_O_NR_002.zip"],  # noqa
                  "target_protocol": "s3",
                  "target_options": {"anon": False,
                                     "client_kwargs": {"endpoint_url": "https://my.dismi.se"}}}
            msg_data = {"dataset": [{"filesystem": fs,
                                     "uid": "zip:///S3A_OL_2_WFR____20201210T080758_20201210T080936_20201210T103707_0097_066_078_1980_MAR_O_NR_002.SEN3/Oa01_reflectance.nc::s3:///sentinel-s3-ol2wfr-zips/2020/12/10/S3A_OL_2_WFR____20201210T080758_20201210T080936_20201210T103707_0097_066_078_1980_MAR_O_NR_002.zip",  # noqa
                                     "uri": "zip://" + filename
                                     }]
                        }

            msg = mock.MagicMock()
            msg.data = msg_data

            prodlist = yaml.load(yaml_test_minimal, Loader=UnsafeLoader)
            jobs = message_to_jobs(msg, prodlist)
            filesystemfile = jobs[999]['input_filenames'][0]

            assert filesystemfile == fsfile.return_value
            fsfile.assert_called_once_with(filename, abs_fs.from_json.return_value)
            abs_fs.from_json.assert_called_once_with(json.dumps(fs))


class TestRun(TestCase):
    """Test case for running the plugins."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.config = yaml.load(yaml_test1, Loader=UnsafeLoader)
        self.queue = multiprocessing.Manager().Queue()

    def test_run_does_not_call_process_directly(self):
        """Test that process is called through Process."""
        from trollflow2.launcher import Runner
        with mock.patch('trollflow2.launcher.yaml.load'),\
                mock.patch('trollflow2.launcher.open'),\
                mock.patch('trollflow2.launcher.generate_messages') as generate_messages,\
                mock.patch('trollflow2.launcher.process') as process,\
                mock.patch('trollflow2.launcher.check_results'),\
                mock.patch('multiprocessing.get_context'):
            generate_messages.side_effect = ['foo', KeyboardInterrupt]
            prod_list = 'bar'
            try:
                runner = Runner(prod_list, self.queue)
                runner.run()
            except KeyboardInterrupt:
                pass
            process.assert_not_called()

    def test_run_uses_process_via_multiprocessing(self):
        """Test that process is called through Process."""
        from trollflow2.launcher import Runner
        from threading import Thread
        with mock.patch('trollflow2.launcher.yaml.load'),\
                mock.patch('trollflow2.launcher.open'),\
                mock.patch('trollflow2.launcher.generate_messages') as generate_messages,\
                mock.patch('trollflow2.launcher.process') as process, \
                mock.patch('multiprocessing.get_context') as get_context:
            def gen_messages(*args):
                del args
                yield 'foo'
                raise KeyboardInterrupt
            generate_messages.side_effect = gen_messages
            get_context.return_value.Process.side_effect = Thread
            prod_list = 'bar'
            try:
                runner = Runner(prod_list, self.queue)
                runner.run()
            except KeyboardInterrupt:
                pass
            process.assert_called_once()

    def test_run_relies_on_listener(self):
        """Test running relies on listener."""
        from trollflow2.launcher import Runner
        with mock.patch('trollflow2.launcher.yaml.load') as yaml_load,\
                mock.patch('trollflow2.launcher.open'),\
                mock.patch('multiprocessing.get_context') as get_context, \
                mock.patch('trollflow2.launcher.ListenerContainer') as lc_:
            listener = mock.MagicMock()
            listener.output_queue.get.return_value = 'foo'
            lc_.return_value = listener
            proc_ret = mock.MagicMock()
            get_context.return_value.Process.return_value = proc_ret
            # stop looping
            proc_ret.join.side_effect = KeyboardInterrupt
            yaml_load.return_value = self.config
            prod_list = 'bar'
            try:
                runner = Runner(prod_list, self.queue)
                runner.run()
            except KeyboardInterrupt:
                pass
            listener.output_queue.called_once()
            lc_.assert_called_with(addresses=None, nameserver='localhost',
                                   topics=['/topic1', '/topic2'])
            # Subscriber topics are removed from config
            self.assertTrue('subscribe_topics' not in self.config['product_list'])
            # Topics are given as command line option
            lc_.reset_mock()
            try:
                runner = Runner(prod_list, self.queue, connection_parameters=dict(topic=['/topic3']))
                runner.run()
            except KeyboardInterrupt:
                pass
            lc_.assert_called_with(addresses=None, nameserver='localhost',
                                   topics=['/topic3'])

    def test_run_starts_and_joins_process(self):
        """Test running."""
        with run_on_a_simple_product_list(self.config, self.queue) as (yaml_load, get_context, lc_, proc_ret):
            proc_ret.start.assert_called_once()
            proc_ret.join.assert_called_once()

    def test_subprocess_is_spawned(self):
        """Test that the subprocess is spawned, not forked."""
        with run_on_a_simple_product_list(self.config, self.queue) as (yaml_load, get_context, lc_, proc_ret):
            get_context.assert_called_once_with("spawn")


@contextmanager
def run_on_a_simple_product_list(config, log_queue):
    """Run a simple (fake) product list."""
    from trollflow2.launcher import Runner
    with mock.patch('trollflow2.launcher.yaml.load') as yaml_load,\
            mock.patch('trollflow2.launcher.open'),\
            mock.patch('multiprocessing.get_context') as get_context,\
            mock.patch('trollflow2.launcher.ListenerContainer') as lc_:
        listener = mock.MagicMock()
        listener.output_queue.get.return_value = 'foo'
        lc_.return_value = listener
        proc_ret = mock.MagicMock()
        get_context.return_value.Process.return_value = proc_ret
        # stop looping
        proc_ret.join.side_effect = KeyboardInterrupt
        yaml_load.return_value = config
        prod_list = 'bar'
        try:
            runner = Runner(prod_list, log_queue)
            runner.run()
        except KeyboardInterrupt:
            pass
        yield yaml_load, get_context, lc_, proc_ret


class TestInterruptRun(TestCase):
    """Test case for running the plugins."""

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.config = yaml.load(yaml_test1, Loader=UnsafeLoader)
        self.queue = multiprocessing.Manager().Queue()

    def test_run_keyboard_interrupt(self):
        """Test interrupting the run with a ctrl-C."""
        from trollflow2.launcher import Runner
        with mock.patch('trollflow2.launcher.yaml.load'),\
                mock.patch('trollflow2.launcher.open'),\
                mock.patch('trollflow2.launcher.ListenerContainer') as lc_:
            listener = mock.MagicMock()
            get = mock.Mock()
            get.side_effect = KeyboardInterrupt
            listener.output_queue.get = get
            lc_.return_value = listener
            runner = Runner(0, self.queue)
            runner.run()
            listener.stop.assert_called_once()


class TestRunLogging(TestCase):
    """Test case for checking the logging in `run`."""

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        """Inject the caplog fixture into this testcase instance."""
        self._caplog = caplog

    def setUp(self):
        """Set up the test case."""
        super().setUp()
        self.config = yaml.load(yaml_test1, Loader=UnsafeLoader)
        self.queue = multiprocessing.Manager().Queue()

    def test_subprocess_uses_queued_logging(self):
        """Test that the subprocess logs are handled."""
        from trollflow2.launcher import Runner
        with mock.patch('trollflow2.launcher.yaml.load'), \
                mock.patch('trollflow2.launcher.open'), \
                mock.patch('trollflow2.launcher.generate_messages') as generate_messages, \
                mock.patch('trollflow2.launcher.process'), \
                mock.patch('trollflow2.launcher.check_results'), \
                mock.patch('multiprocessing.get_context'):
            generate_messages.side_effect = ['foo', KeyboardInterrupt]
            prod_list = 'bar'
            try:
                runner = Runner(prod_list, self.queue)
                runner.run()
            except KeyboardInterrupt:
                pass
            root_logger = logging.getLogger()
            for handler in root_logger.handlers:
                if isinstance(handler, QueueHandler):
                    break
            else:
                raise AssertionError("No QueueHandler found.")


class TestExpand(TestCase):
    """Test expanding the product list."""

    def test_expand(self):
        """Test expanding the product list."""
        from trollflow2.launcher import expand
        inside = {'a': 'b'}
        outside = {'c': inside, 'd': inside}
        expanded = expand(outside)
        self.assertIsNot(expanded['d'], expanded['c'])


class TestProcess(TestCase):
    """Test the process function."""

    def setUp(self):
        """Set up the test."""
        self.patcher = mock.patch.multiple("trollflow2.launcher",
                                           traceback=mock.DEFAULT,
                                           sendmail=mock.DEFAULT,
                                           expand=mock.DEFAULT,
                                           yaml=mock.DEFAULT,
                                           message_to_jobs=mock.DEFAULT,
                                           open=mock.DEFAULT,
                                           get_dask_client=mock.DEFAULT)
        mocks = self.patcher.start()

        self.traceback = mocks['traceback']
        self.sendmail = mocks['sendmail']

        self.open = mocks['open']
        fid = mock.MagicMock()
        fid.read.return_value = yaml_test1
        self.open.return_value.__enter__.return_value = fid

        self.yaml = mocks['yaml']
        # Is this necessary?
        self.yaml.YAMLError = YAMLError

        self.get_dask_client = mocks['get_dask_client']
        # Make a client that has no `.close()` method (for coverage)
        self.client = mock.MagicMock()
        self.client.close.side_effect = AttributeError
        self.get_dask_client.return_value = self.client

        self.expand = mocks['expand']
        self.fake_plugin = mock.MagicMock()
        # Return something resembling a config
        self.expand.return_value = {"workers": [{"fun": self.fake_plugin}]}

        self.message_to_jobs = mocks['message_to_jobs']
        self.message_to_jobs.return_value = {1: {"job1": dict([])}}

        self.queue = mock.MagicMock()

    def tearDown(self):
        """Tear down the test case."""
        self.patcher.stop()

    def test_plugin_is_stopped_after_processing(self):
        """Test plugin is stopped after processing."""
        self.fake_plugin.stop.assert_not_called()
        process("msg", "prod_list", self.queue)
        self.fake_plugin.stop.assert_called_once()

    def test_product_list_is_opened(self):
        """Test product list is opened."""
        process("msg", "prod_list", self.queue)
        self.open.assert_called_with("prod_list")

    def test_yaml_config_is_read_only_once(self):
        """Test that the yaml config is read only once."""
        process("msg", "prod_list", self.queue)
        self.yaml.load.assert_called_once()

    def test_workers_config_is_passed_down(self):
        """Test that the workers config is used."""
        process("msg", "prod_list", self.queue)
        self.message_to_jobs.assert_called_with("msg", {"workers": [{"fun": self.fake_plugin}]})

    def test_plugin_is_used(self):
        """Test that the plugin is being used."""
        process("msg", "prod_list", self.queue)
        self.fake_plugin.assert_called_with({'job1': {}, 'processing_priority': 1, 'produced_files': self.queue})

    def test_dask_client_is_used(self):
        """Test that the dask client is used."""
        process("msg", "prod_list", self.queue)
        self.get_dask_client.assert_called_once()

    def test_dask_client_is_closed(self):
        """Test that the dask client is closed."""
        process("msg", "prod_list", self.queue)
        self.client.close.assert_called_once()

    def test_plugin_with_no_stop_work(self):
        """Test that plugins with no `stop` method (like regular functions) can be used."""
        self.fake_plugin.stop = mock.MagicMock(side_effect=AttributeError('boo'))
        process("msg", "prod_list", self.queue)

    def test_error_propagation(self):
        """Test that errors are propagated."""
        self.fake_plugin.side_effect = KeyboardInterrupt
        with pytest.raises(KeyboardInterrupt):
            process("msg", "prod_list", self.queue)

    def test_crash_handler_call(self):
        """Test crash hander call.

        This will raise KeyError as there are no configured workers in the config returned by expand().
        """
        self.traceback.format_exc.return_value = 'baz'
        crash_handlers = {"crash_handlers": {"config": {"foo": "bar"},
                                             "handlers": [{"fun": self.sendmail}]}}
        self.expand.return_value = crash_handlers
        with pytest.raises(KeyError):
            process("msg", "prod_list", self.queue)
        config = crash_handlers['crash_handlers']['config']
        self.sendmail.assert_called_once_with(config, 'baz')

    def test_open_missing_file(self):
        """Test failure in open() due to a missing config file."""
        self.open.side_effect = IOError
        with pytest.raises(IOError):
            process("msg", "prod_list", self.queue)

    def test_open_bad_yaml(self):
        """Test failure in yaml.load(), e.g. bad formatting."""
        self.open.side_effect = YAMLError
        with pytest.raises(YAMLError):
            process("msg", "prod_list", self.queue)

    @pytest.mark.skipif(sys.platform != "linux",
                        reason="Timeout only supported on Linux")
    def test_timeout_in_running_job(self):
        """Test timeout in running job."""
        def wait(job):
            del job
            time.sleep(0.1)

        self.fake_plugin.side_effect = wait

        self.expand.return_value = {"workers": [{"fun": self.fake_plugin, "timeout": 0.05}]}
        with pytest.raises(TimeoutError, match="Timeout for .* expired "
                                               "after 0.1 seconds"):
            process("msg", "prod_list", self.queue)
        # wait a little to ensure alarm is not raised later
        time.sleep(0.11)


def test_workers_initialized():
    """Test that the config loading works when workers are defined."""
    from tempfile import NamedTemporaryFile
    import os

    queue = mock.MagicMock()

    with NamedTemporaryFile(mode='w+t', delete=False) as tmp_file:
        fname = tmp_file.name
        tmp_file.write(yaml_test_minimal)
        tmp_file.close()

        try:
            with mock.patch("trollflow2.launcher.get_dask_client") as gdc:
                # `get_dask_client()` is called just after config reading, so if we get there loading worked
                gdc.side_effect = StopIteration
                try:
                    process("msg", fname, queue)
                except StopIteration:
                    pass
        finally:
            os.remove(fname)


def test_get_dask_client(caplog):
    """Test getting dask client."""
    from trollflow2.launcher import get_dask_client

    ncores = mock.MagicMock()
    ncores.return_value = {}
    client = mock.MagicMock(ncores=ncores)
    client_class = mock.MagicMock()
    client_class.return_value = client

    # No client configured
    config = {}
    with caplog.at_level(logging.DEBUG):
        res = get_dask_client(config)
    assert "Distributed processing not configured" in caplog.text
    caplog.clear()
    assert res is None

    # Config is valid, but no workers are available
    config = {"dask_distributed": {"class": client_class,
                                   "settings": {"foo": 1, "bar": 2}
                                   }
              }
    with caplog.at_level(logging.WARNING):
        res = get_dask_client(config)
    assert "No workers available, reverting to default scheduler" in caplog.text
    caplog.clear()
    assert res is None
    ncores.assert_called_once()
    client.close.assert_called_once()

    # The scheduler had no workers, the client doesn't have `.close()`
    client.close.side_effect = AttributeError
    with caplog.at_level(logging.WARNING):
        res = get_dask_client(config)
    assert res is None

    # Config is valid, scheduler has workers
    ncores.return_value = {'a': 1, 'b': 1}
    with caplog.at_level(logging.DEBUG):
        res = get_dask_client(config)
    assert "Using dask distributed client" in caplog.text
    caplog.clear()
    assert res is client
    assert ncores.call_count == 3

    # Scheduler couldn't connect to workers
    client_class.side_effect = OSError
    with caplog.at_level(logging.ERROR):
        res = get_dask_client(config)
    assert "Scheduler not found, reverting to default scheduler" in caplog.text
    caplog.clear()
    assert res is None
    assert ncores.call_count == 3


def test_check_results(tmp_path, caplog):
    """Test functionality for check_results."""
    from trollflow2.launcher import check_results

    class FakeQueue:
        def __init__(self, lo, hi, skip=None):
            if skip is None:
                skip = []
            self._files = set()
            for i in range(lo, hi):
                f = (tmp_path / f"file{i:d}")
                self._files.add(str(f))
                if i not in skip:
                    with f.open(mode="wt") as fp:
                        fp.write("zucchini" * i)

        def get(self, block=None):
            try:
                return self._files.pop()
            except KeyError:
                raise queue.Empty

        def qsize(self):
            return len(self._files)

    produced_files = FakeQueue(0, 3)
    start_time = datetime.datetime(1900, 1, 1)
    exitcode = 0
    with caplog.at_level(logging.DEBUG):
        check_results(produced_files, start_time, exitcode)
    assert "Empty file detected" in caplog.text
    assert "files produced nominally" not in caplog.text

    produced_files = FakeQueue(5, 8, skip=[6])
    with caplog.at_level(logging.DEBUG):
        check_results(produced_files, start_time, exitcode)
    assert "Missing file" in caplog.text
    assert "files produced nominally" not in caplog.text

    produced_files = FakeQueue(10, 13)
    with caplog.at_level(logging.DEBUG), \
            mock.patch("trollflow2.launcher.datetime") as dd:
        dd.now.return_value = datetime.datetime(1927, 5, 20, 0, 0)
        check_results(produced_files, start_time, exitcode)
    assert "All 3 files produced nominally in 10000 days" in caplog.text

    with caplog.at_level(logging.DEBUG):
        check_results(produced_files, start_time, 1)
    assert "Process crashed with exit code 1" in caplog.text

    with caplog.at_level(logging.DEBUG):
        check_results(produced_files, start_time, -1)
    assert "Process killed with signal 1" in caplog.text


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Pytroll developers
#
# Author(s):
#
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
"""The launcher module.

This delegate the actual running of the plugins to a subprocess to avoid any
memory buildup.
"""


import ast
import copy
import gc
import os
import re
import signal
import traceback
from collections import OrderedDict
from contextlib import suppress
from datetime import datetime
from logging import getLogger
from queue import Empty
from urllib.parse import urlparse

import yaml
from trollflow2.dict_tools import gen_dict_extract, plist_iter
from trollflow2.logging import setup_queued_logging
from trollflow2.plugins import AbortProcessing

try:
    from posttroll.listener import ListenerContainer
except ImportError:
    ListenerContainer = None

from yaml import UnsafeLoader, SafeLoader, BaseLoader


LOG = getLogger(__name__)
DEFAULT_PRIORITY = 999


def tuple_constructor(loader, node):
    """Construct a tuple."""
    def parse_tup_el(el):
        return ast.literal_eval(el.strip())
    value = loader.construct_scalar(node)
    tup_elements = value[1:-1].split(',')
    if tup_elements[-1] == '':
        tup_elements.pop(-1)
    tup = tuple((parse_tup_el(el) for el in tup_elements))
    return tup


tuple_regex = r'\( *([\w.]+|"[\w\s.]*") *(, *([\w.]+|"[\w\s.]*") *)*((, *([\w.]+|"[\w\s.]*") *)|(, *))\)'
yaml.add_constructor(u'!tuple', tuple_constructor, UnsafeLoader)
yaml.add_implicit_resolver(u'!tuple', re.compile(tuple_regex), None, UnsafeLoader)


def get_test_message(test_message_file):
    """Read file and retrieve the test message."""
    msg = None
    if test_message_file:
        with open(test_message_file) as fpt:
            msg = fpt.readline().strip('\n')

    return msg


def check_results(produced_files, start_time, exitcode):
    """Make sure the composites have been saved."""
    end_time = datetime.now()
    error_detected = False
    try:
        qsize = produced_files.qsize()
    except NotImplementedError:  # MacOS
        qsize = None
    while True:
        try:
            saved_file = produced_files.get(block=False)
            try:
                if os.path.getsize(saved_file) == 0:
                    LOG.error("Empty file detected: %s", saved_file)
                    error_detected = True
            except FileNotFoundError:
                LOG.error("Missing file: %s", saved_file)
                error_detected = True
        except Empty:
            break
    if exitcode != 0:
        error_detected = True
        if exitcode < 0:
            LOG.error('Process killed with signal %d', -exitcode)
        else:
            LOG.critical('Process crashed with exit code %d', exitcode)
    if not error_detected:
        elapsed = end_time - start_time
        if qsize is not None:
            LOG.info(f'All {qsize:d} files produced nominally in '
                     f"{elapsed!s}", extra={"time": elapsed})
        else:
            LOG.info(f'All files produced nominally in '
                     f"{elapsed!s}", extra={"time": elapsed})


def generate_messages(connection_parameters):
    """Generate messages using a ListenerContainer."""
    listener = _create_listener_from_connection_parameters(connection_parameters)
    while True:
        try:
            yield listener.output_queue.get(True, 5)
        except KeyboardInterrupt:
            listener.stop()
            return
        except Empty:
            continue


def _create_listener_from_connection_parameters(connection_parameters):
    """Create listener from connection parameters."""
    topics = connection_parameters['topic']
    nameserver = connection_parameters.get('nameserver', 'localhost')
    addresses = connection_parameters.get('addresses')
    listener = ListenerContainer(
        addresses=addresses,
        nameserver=nameserver,
        topics=topics)
    return listener


class Runner:
    """Class that handles all the administration around running on a product list."""

    def __init__(self, product_list, log_queue, connection_parameters=None, test_message=None, threaded=False):
        """Set up the runner."""
        self.product_list = product_list
        self.log_queue = log_queue
        self.connection_parameters = connection_parameters
        self.test_message = get_test_message(test_message)
        self.threaded = threaded

    def run(self):
        """Spawn one or multiple subprocesses or threads to run the jobs from the product list."""
        messages = self._get_message_iterator()

        if self.threaded:
            self._run_threaded(messages)
        else:
            self._run_subprocess(messages)

    def _get_message_iterator(self):
        """Get the messages to work on."""
        if self.test_message:
            from posttroll.message import Message
            messages = [Message(rawstr=self.test_message)]
            self.threaded = True
        else:
            self._fill_in_connection_parameters()
            messages = generate_messages(self.connection_parameters)
        return messages

    def _fill_in_connection_parameters(self):
        """Fill in the connection parameters for the message listener."""
        with open(self.product_list) as fid:
            config = yaml.load(fid.read(), Loader=BaseLoader)
        if self.connection_parameters is None:
            self.connection_parameters = dict()
        if not self.connection_parameters.get('topic'):
            self.connection_parameters['topic'] = config['product_list'].pop('subscribe_topics', None)

    def _run_threaded(self, messages):
        """Run in a thread."""
        LOG.info("Launching trollflow2 with threads")
        from threading import Thread
        self._run_product_list_on_messages(messages, process, Thread)

    def _run_subprocess(self, messages):
        """Run in a subprocess, with queued logging."""
        LOG.info("Launching trollflow2 with subprocesses")
        from multiprocessing import get_context
        ctx = get_context("spawn")
        self._run_product_list_on_messages(messages, queue_logged_process, ctx.Process)

    def _run_product_list_on_messages(self, messages, target_fun, process_class):
        """Run the product list on the messages."""
        for msg in messages:
            produced_files_queue = self.log_queue._manager.Queue()
            kwargs = dict(produced_files=produced_files_queue, prod_list=self.product_list)
            if not self.threaded:
                kwargs["log_queue"] = self.log_queue
            proc = process_class(target=target_fun, args=(msg,), kwargs=kwargs)
            start_time = datetime.now()
            proc.start()
            proc.join()
            try:
                exitcode = proc.exitcode
            except AttributeError:
                exitcode = 0
            check_results(produced_files_queue, start_time, exitcode)


def get_area_priorities(product_list):
    """Get processing priorities and names for areas."""
    priorities = {}
    plist = product_list['product_list']['areas']
    for area in plist.keys():
        prio = plist[area].get('priority', DEFAULT_PRIORITY)
        if prio is None:
            prio = DEFAULT_PRIORITY
        if prio not in priorities:
            priorities[prio] = [area]
        else:
            priorities[prio].append(area)

    return priorities


def message_to_jobs(msg, product_list):
    """Convert a posttroll message *msg* to a list of jobs given a *product_list*."""
    formats = product_list['product_list'].get('formats', None)
    for _product, pconfig in plist_iter(product_list['product_list'], level='product'):
        if 'formats' not in pconfig and formats is not None:
            pconfig['formats'] = copy.deepcopy(formats)
    jobs = OrderedDict()
    priorities = get_area_priorities(product_list)
    # TODO: check the uri is accessible from the current host.
    input_filenames = _extract_filenames(msg)
    for prio, areas in priorities.items():
        jobs[prio] = OrderedDict()
        jobs[prio]['input_filenames'] = input_filenames.copy()
        jobs[prio]['input_mda'] = msg.data.copy()
        jobs[prio]['product_list'] = {}
        for section in product_list:
            if section == 'product_list':
                if section not in jobs[prio]['product_list']:
                    jobs[prio]['product_list'][section] = OrderedDict(product_list[section].copy())
                    del jobs[prio]['product_list'][section]['areas']
                    jobs[prio]['product_list'][section]['areas'] = OrderedDict()
                for area in areas:
                    jobs[prio]['product_list'][section]['areas'][area] = product_list[section]['areas'][area]
            else:
                jobs[prio]['product_list'][section] = product_list[section]
    return jobs


def _extract_filenames(msg):
    """Extract the filenames from *msg*.

    If the message contains a `filesystem` item, use fsspec to decode it.
    """
    filenames = [urlparse(uri).path for uri in gen_dict_extract(msg.data, 'uri')]
    filenames = _create_fs_file_instances(filenames, msg)
    return filenames


def _create_fs_file_instances(filenames, msg):
    """Create FSFile instances when filesystem is provided."""
    filesystems = list(gen_dict_extract(msg.data, 'filesystem'))
    if filesystems:
        from satpy.readers import FSFile
        from fsspec.spec import AbstractFileSystem
        import json
        filenames = [FSFile(filename, AbstractFileSystem.from_json(json.dumps(filesystem)))
                     for filename, filesystem in zip(filenames, filesystems)]
    return filenames


def expand(yml):
    """Expand a yaml config so that aliases are copied.

    PFE http://disq.us/p/1tdbxgx
    """
    if isinstance(yml, dict):
        for key, value in yml.items():
            if isinstance(value, dict):
                expand(value)
                yml[key] = copy.deepcopy(yml[key])
    return yml


def get_dask_client(config):
    """Create Dask client if configured."""
    client = None

    try:
        client_class = config["dask_distributed"]["class"]
        settings = config["dask_distributed"].get("settings", {})
        client = client_class(**settings)
        try:
            if not client.ncores():
                LOG.warning("No workers available, reverting to default scheduler")
                client.close()
                client = None
        except AttributeError:
            client = None
    except OSError:
        LOG.error("Scheduler not found, reverting to default scheduler")
    except KeyError:
        LOG.debug("Distributed processing not configured, "
                  "using default scheduler")
    else:
        LOG.debug(f"Using dask distributed client {client!s}")

    return client


def queue_logged_process(msg, prod_list, produced_files, log_queue):
    """Run `process` with a queued log."""
    setup_queued_logging(log_queue)
    with suppress(ValueError):
        signal.signal(signal.SIGUSR1, print_traces)
        LOG.debug("Use SIGUSR1 on pid {} to check the current tracebacks of this subprocess.".format(os.getpid()))
    process(msg, prod_list, produced_files)


def print_traces(signum, frame):
    """Print traces for debugging."""
    import sys
    import traceback
    print(f"Got signal {signum} in {frame}, dumping traces.", file=sys.stderr)

    for thread, current_frame in sys._current_frames().items():
        print('Thread 0x%x' % thread, file=sys.stderr)
        traceback.print_stack(current_frame, file=sys.stderr)
        print(file=sys.stderr)


def process(msg, prod_list, produced_files):
    """Process a message."""
    config = read_config(prod_list, Loader=UnsafeLoader)

    # Get distributed client
    client = get_dask_client(config)

    try:
        config = expand(config)
        jobs = message_to_jobs(msg, config)
        for prio in sorted(jobs.keys()):
            job = jobs[prio]
            job['processing_priority'] = prio
            job['produced_files'] = produced_files
            try:
                for wrk in config['workers']:
                    cwrk = wrk.copy()
                    if "timeout" in cwrk:

                        def _timeout_handler(signum, frame, wrk=wrk):
                            raise TimeoutError(
                                f"Timeout for {wrk['fun']!s} expired "
                                f"after {wrk['timeout']:.1f} seconds, "
                                "giving up")
                        signal.signal(signal.SIGALRM, _timeout_handler)
                        # using setitimer because it accepts floats,
                        # unlike signal.alarm
                        signal.setitimer(signal.ITIMER_REAL,
                                         cwrk.pop("timeout"))
                    cwrk.pop('fun')(job, **cwrk)
                    if "timeout" in cwrk:
                        signal.alarm(0)  # cancel the alarm
            except AbortProcessing as err:
                LOG.warning(str(err))
    except Exception:
        LOG.exception("Process crashed")
        if "crash_handlers" in config:
            trace = traceback.format_exc()
            for hand in config['crash_handlers']['handlers']:
                hand['fun'](config['crash_handlers']['config'], trace)
        raise
    finally:
        # Remove config and run garbage collection so all remaining
        # references e.g. to FilePublisher should be removed
        LOG.debug('Cleaning up')
        for wrk in config.get("workers", []):
            try:
                wrk['fun'].stop()
            except AttributeError:
                continue
        del config
        try:
            client.close()
        except AttributeError:
            pass
        gc.collect()


def read_config(fname=None, raw_string=None, Loader=SafeLoader):
    """Read the configuration file."""
    try:
        if fname:
            with open(fname) as fid:
                raw_config = fid.read()
        elif raw_string:
            raw_config = raw_string
        config = yaml.load(_remove_null_keys(raw_config), Loader=Loader)
    except (IOError, yaml.YAMLError):
        # Either open() or yaml.load() failed
        LOG.exception("Process crashed, check YAML file.")
        raise
    return config


def _remove_null_keys(raw_config):
    return raw_config.replace('null:', 'None:')


def sendmail(config, trace):
    """Send email about crashes using `sendmail`."""
    from email.mime.text import MIMEText
    from subprocess import Popen, PIPE

    email_settings = config['sendmail']
    msg = MIMEText(email_settings["header"] + "\n\n" + "\n\n" + trace)
    msg["From"] = email_settings["from"]
    msg["To"] = email_settings["to"]
    msg["Subject"] = email_settings["subject"]
    sendmail = email_settings.get("sendmail", "/usr/bin/sendmail")

    pid = Popen([sendmail, "-t", "-oi"], stdin=PIPE)
    pid.communicate(msg.as_bytes())
    pid.terminate()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pytroll developers
#
# Author(s):
#
#   Martin Raspaud <martin.raspaud@smhi.se>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
"""Logging utilities."""

import logging
import logging.config
from contextlib import contextmanager
from logging import getLogger, DEBUG
from logging.handlers import QueueListener, QueueHandler

DEFAULT_LOG_CONFIG = {'version': 1,
                      'disable_existing_loggers': False,
                      'formatters': {'pytroll': {'format': '[%(levelname)s: %(asctime)s : %(name)s] %(message)s',
                                                 'datefmt': '%Y-%m-%d %H:%M:%S'}},
                      'handlers': {'console': {'class': 'logging.StreamHandler',
                                               'formatter': 'pytroll'}},
                      'root': {'level': 'DEBUG', 'handlers': ['console']}}


@contextmanager
def logging_on(log_queue, config=None):
    """Activate queued logging.

    This context activates logging through the use of logging's QueueHandler and
    QueueListener.
    Whether the default config parameters are used or a custom configuration is
    passed, the log handlers are passed to a QueueListener instance, such that
    the subprocesses of trollflow2 can use a QueueHandler to communicate logs.
    """
    root = logging.getLogger()
    # Lift out the existing handlers (we need to keep these for pytest's caplog)
    handlers = root.handlers.copy()

    if config is None:
        config = DEFAULT_LOG_CONFIG
    logging.config.dictConfig(config)

    # set up and run listener
    listener = QueueListener(log_queue, *(root.handlers + handlers))
    listener.start()
    try:
        yield
    finally:
        listener.stop()


def setup_queued_logging(log_queue):
    """Set up queued logging."""
    root_logger = getLogger()
    root_logger.addHandler(QueueHandler(log_queue))
    root_logger.setLevel(DEBUG)
